/*****************************************************************************
 *
 * FILE:	cond.c
 * DESCRIPTION:	addressing the conditional request
 * DATE:	Wed, Dec 15 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cond.c,v 1.15.2.4 2003/11/28 15:59:58 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include "wyvern.h"
#include "functions.h"



/*****************************************************************************
 *
 *	declaration of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	ifModifiedSince(Thread *);
static void	ifUnmodifiedSince(Thread *);
static void	ifMatch(Thread *);
static void	ifNoneMatch(Thread *);
static void	ifRange(Thread *);
static void	getRequestByteRanges(Thread *);
static void	setContentRange(Thread *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	checkConditionalRequest
 * Description:		check the conditional request
 * Arguments:		thread
 * Return Value:	none
 */
void
checkConditionalRequest(thd)
	register Thread *	thd;
{
  /* set the entity tag for ETag/If-Match/If-None-Match/If-Range */
  setETag(thd);

  if (thd->request.if_range != NULL) {
    ifRange(thd);
  }
  else if (thd->request.if_modified_since != NULL) {
    ifModifiedSince(thd);
  }
  else if (thd->request.if_unmodified_since != NULL) {
    ifUnmodifiedSince(thd);
  }
  else if (thd->request.if_match != NULL) {
    ifMatch(thd);
  }
  else if (thd->request.if_none_match != NULL) {
    ifNoneMatch(thd);
  }
  else {
    setResponseCode(thd, CODE_REQUEST_OK);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	ifModifiedSince
 * Description:		check the conditional request with If-Modified-Since
 * Arguments:		thread
 * Return Value:	none
 */
static void
ifModifiedSince(thd)
	register Thread *	thd;
{
  time_t	ims;

  ims = convertDate(thd->request.if_modified_since);
  if (ims == (time_t)0) {
    setResponseCode(thd, CODE_FORBIDDEN);
  }
  else if (thd->file.st_mtime <= ims) {
    setResponseCode(thd, CODE_NOT_MODIFIED);
  }
}

/*
 * Function Name:	ifUnmodifiedSince
 * Description:		check the conditional request with If-Unmodified-Since
 * Arguments:		thread
 * Return Value:	none
 */
static void
ifUnmodifiedSince(thd)
	register Thread *	thd;
{
  time_t	ius;

  ius = convertDate(thd->request.if_unmodified_since);
  if (ius == (time_t)0) {
    setResponseCode(thd, CODE_FORBIDDEN);
  }
  else if (thd->file.st_mtime > ius) {
    setResponseCode(thd, CODE_PRECONDITION_FAILED);
  }
}

/*
 * Function Name:	ifMatch
 * Description:		check the conditional request with If-Match
 * Arguments:		thread
 * Return Value:	none
 */
static void
ifMatch(thd)
	register Thread *	thd;
{
  if (strcmp(thd->request.if_match, thd->response.etag) != 0 &&
      strcmp(thd->request.if_match, "*") != 0) {
    setResponseCode(thd, CODE_PRECONDITION_FAILED);
  }
}

/*
 * Function Name:	ifNoneMatch
 * Description:		check the conditional request with If-None-Match
 * Arguments:		thread
 * Return Value:	none
 */
static void
ifNoneMatch(thd)
	register Thread *	thd;
{
  if (strcmp(thd->request.if_none_match, thd->response.etag) == 0 ||
      strcmp(thd->request.if_none_match, "*") == 0) {
    setResponseCode(thd, CODE_PRECONDITION_FAILED);
  }
}

/*
 * Function Name:	ifRange
 * Description:		check the conditional request with If-Range
 * Arguments:		thread
 * Return Value:	none
 */
static void
ifRange(thd)
	register Thread *	thd;
{
  if (thd->request.range == NULL) {
    return;
  }

  /* check the entity-tag */
  if (strcmp(thd->request.if_range, thd->response.etag) == 0) {
    getRequestByteRanges(thd);
  }
  else {
    /* check the HTTP-date */
    time_t	t = convertDate(thd->request.if_range);
    if (thd->file.st_mtime == t) {
      getRequestByteRanges(thd);
    }
  }
}

/*
 * Function Name:	getRequestByteRanges
 * Description:		get the byte ranges
 * Arguments:		thread
 * Return Value:	none
 */
static void
getRequestByteRanges(thd)
	register Thread *	thd;
{
  String	p;
  String	first;
  String	last;

  p = strstr(thd->request.range, "bytes=");
  if (p == NULL) {
    return;	/* invalid range specification */
  }
  first = p + 6;

  p = strchr(first, '-');
  if (p == NULL) {
    return;	/* invalid range specification */
  }
  *p++ = EOL;
  last = p;

  if (*first == '-') {		/* ex. bytes=-500 */
    thd->range.first = 0;
    thd->range.last  = atoi(last);
  }
  else if (*first == EOL && *last == EOL) {
    return;	/* invalid range specification */
  }
  else if (*last == EOL) {	/* ex. bytes=9500- */
    thd->range.first = atoi(first);
    thd->range.last  = thd->file.st_size;
  }
  else {			/* ex. bytes=500-900 */
    thd->range.first = atoi(first);
    thd->range.last  = atoi(last);
  }

  /* set file offset */
  if (fseek(thd->fp, (long)thd->range.first, SEEK_SET) == -1) {
    thd->range.first = thd->range.last = 0;
    return;
  }

  setResponseCode(thd, CODE_PARTIAL_CONTENT);
  setContentRange(thd);
}

/*
 * Function Name:	setContentRange
 * Description:		set the Content-Range
 * Arguments:		thread
 * Return Value:	none
 */
static void
setContentRange(thd)
	register Thread *	thd;
{
  if (xasprintf(&thd->response.content_range, "%s %d-%d/%ld",
		RANGE_UNIT,
		thd->range.first, thd->range.last,
		(long)thd->file.st_size) == -1) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
  }
}

