/*****************************************************************************
 *
 * FILE:	config.c
 * DESCRIPTION:	read configuration file
 * DATE:	Wed, Sep  8 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: config.c,v 1.26.2.3 2002/08/10 20:49:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/
static void	verifyConfiguration(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	readConfigFile
 * Description:		Reads config file via yyparse.
 * Arguments:		none
 * Return Value:	none
 */
void
readConfigFile(void)
{
  extern FILE *		yyin;
  extern int		yyparse(void);
  struct stat		sbuf;

  /* change working directory */
  if (chdir(SERVER_ROOT) == -1) {
    die("failed to change directory: %s\n", SERVER_ROOT);
  }

  /* open config file */
  if (lstat(httpd->configFile, &sbuf) == 0) {
    if (S_ISLNK(sbuf.st_mode)) {
      die("%s: untrusted symbolic link\n", httpd->configFile);
    }

    yyin = fopen(httpd->configFile, "r");
    if (yyin == NULL) {
      die("failed to open %s\n", httpd->configFile);
    }

    /*
     * yyparse() returns
     *	 0: if parsing is successful
     *	 1: if parsing failed
     */
    if (yyparse()) {
      die("failed to parse %s\n", httpd->configFile);
    }

    /* verify configuration and virtual hosts */
    verifyConfiguration();
    verifyVirtualHosts();
  }
  else {
    switch (errno) {
      case ENOENT:
	die("%s doesn't exist\n", httpd->configFile);
      case EACCES:
	die("failed to access %s\n", httpd->configFile);
      default:
	die("failed to open %s\n", httpd->configFile);
    }
  }
}

/*
 * Function Name:	verifyConfiguration
 * Description:		verify configuration of wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
verifyConfiguration(void)
{
  /* don't allow to set root id */
  if (httpd->uid == UID_ROOT) {	
    httpd->uid = UID_NOBODY;
  }

  /* don't allow to set root group id */
  if (httpd->gid == UID_WHEEL) {
    httpd->gid = UID_NOGROUP;
  }

  if (httpd->serverRoot == NULL) {
    httpd->serverRoot = xstrdup(SERVER_ROOT);
  }

  if (httpd->pidFile == NULL) {
    httpd->pidFile = xstrdup(DEF_PID_FILE);
  }

  if (httpd->host->serverName == NULL) {
    httpd->host->serverName = getServerName();
  }

  if (httpd->host->serverAdmin == NULL) {
    die("No ServerAdmin in wyvern.conf\n");
  }
  if (httpd->host->docRoot == NULL) {
    die("No DocRoot in wyvern.conf\n");
  }
  if (httpd->host->errorLog == NULL) {
    die("No ErrorLog in wyvern.conf\n");
  }
  if (httpd->host->accessLog == NULL) {
    die("No AccessLog in wyvern.conf\n");
  }

  if (httpd->host->dirIndex == NULL) {
    httpd->host->dirIndex = xstrdup(DEF_DIRINDEX);	/* DirectoryIndex */
  }

  if (httpd->host->defaultType == NULL) {
    httpd->host->defaultType = xstrdup(DEF_DEFTYPE);	/* DefaultType */
  }
}

/*
 * Function Name:	Reconfiguration
 * Description:		Read wyvern config file again and restart
 * Arguments:		none
 * Return Value:	none
 */
void
Reconfiguration(void)
{
  extern FILE *		yyin;
  extern int		yyparse(void);
  sigset_t		set;
  sigset_t		oset;


  /* block all signals */
  sigfillset(&set);
  sigprocmask(SIG_BLOCK, &set, &oset);

  /*
   * start new configuration
   */

  /* address current request in thread queue */
  flushThreadQueue();

  /* open config file */
  yyin = fopen(httpd->configFile, "r");
  if (yyin == NULL) {
    wvSyslog(LOG_ERR, "failed to open wyvern.conf\n");
    goto end;
  }

  /* unload all modules */
  cleanupModules();

  /* release used memory area */
  xfree((String)httpd->serverRoot);	/* Is this good? */
  xfree((String)httpd->pidFile);	/* Is this good? */
  closeLogFiles();			/* close log files */
  freeErrorDocumentTable();
  freeMimeTypeTable();
  deleteVirtualHosts();

  httpd->vhostList = newVirtualHost(NULL);	/* init VirtualHost queue */
  httpd->vhost	   = httpd->vhostList;		/* current virtual host */
  httpd->host	   = httpd->vhostList;		/* original server host */

  /*
   * yyparse() returns
   *	0: if parsing is successful
   *	1: if parsing failed
   */
  if (yyparse()) {
    die("failed to parse wyvern.conf\n");
  }

  /* verify configuration */
  verifyConfiguration();
  verifyVirtualHosts();

  /* close log files and open logs file again */
  openLogFiles();

  /* reconfiguration finished. */
  wvSyslog(LOG_NOTICE, "%s reconfigured --- resuming normal operations\n",
	   SERVER_VERSION);

end:
  /* restore signal mask */
  sigprocmask(SIG_SETMASK, &oset, NULL);

  /* reset reboot flag */
  httpd->reboot = FALSE;
}

