/*****************************************************************************
 *
 * FILE:	error.c
 * DESCRIPTION:	send an error message
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: error.c,v 1.35.2.3 2002/01/30 13:26:33 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <stdlib.h>
#include <string.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
typedef struct _errorDocument {
  int		code;	/* response status code */
  String	text;	/* document to response */
  struct _errorDocument *	next;
} errorDocument;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static errorDocument *	errorDocumentTable[ERRDOC_TABLE_SIZE];


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static unsigned int	hash(int);
static bool_t		sendCustomizedErrorMessage(Thread *);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	initErrorDocumentTable
 * Description:		initialize hash table for error document
 * Arguments:		none
 * Return Value:	none
 */
void
initErrorDocumentTable(void)
{
  register int	i;

  for (i = 0; i < ERRDOC_TABLE_SIZE; i++) {
    errorDocumentTable[i] = NULL;
  }
}

/*
 * Function Name:	freeErrorDocumentTable
 * Description:		free the memory area of error document table
 * Arguments:		none
 * Return Value:	none
 */
void
freeErrorDocumentTable(void)
{
  register int	i;

  for (i = 0; i < ERRDOC_TABLE_SIZE; i++) {
    register errorDocument *	edp;
    register errorDocument *	tmp;

    if (errorDocumentTable[i] == NULL) {
      continue;
    }
    for (edp = errorDocumentTable[i]; edp != NULL; edp = tmp) {
      tmp = edp->next;
      XFree((String)edp->text);
      XFree((errorDocument *)edp);
    }
    errorDocumentTable[i] = NULL;
  }
}

/*
 * Function Name:	registerErrorDocument
 * Description:		register response code and it's document file to hash
 * Arguments:		response code, document location
 * Return Value:	none
 */
void
registerErrorDocument(code, text)
	int	code;
	String	text;
{
  unsigned int		h;
  errorDocument *	edp;

  if (getErrorDocument(code) != NULL) {
    return;	/* already registered */
  }

  edp = (errorDocument *)xcalloc(1, sizeof(errorDocument));
  if (edp == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }

  h			= hash(code);
  edp->code		= code;
  edp->text		= xstrdup(text);
  edp->next		= errorDocumentTable[h];
  errorDocumentTable[h] = edp;
}

/*
 * Function Name:	getErrorDocument
 * Description:		get error document for response status code
 * Arguments:		response status code
 * Return Value:	pointer of error document
 */
String
getErrorDocument(code)
	int	code;
{
  register errorDocument *	edp;

  for (edp = errorDocumentTable[hash(code)]; edp != NULL; edp = edp->next) {
    if (edp->code == code) {
      return edp->text;
    }
  }

  return NULL;
}

/*
 * Function Name:	hash
 * Description:		calculates a hash value of number
 * Arguments:		number
 * Return Value:	hash value
 */
static unsigned int
hash(code)
	int	code;
{
  return (unsigned int)((67 * code) % ERRDOC_TABLE_SIZE);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	sendErrorMessage
 * Description:		send an error message
 * Arguments:		thread
 * Return Value:	none
 */
void
sendErrorMessage(thd)
	register Thread *	thd;
{
  if (sendCustomizedErrorMessage(thd)) {
    return;
  }

  /*
   * send header part
   */
  iosPrintf(thd,
	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n"
	"<html lang=\"ja\">\n"
	"<head>\n"
	"<style type=\"text/css\">\n"
	"<!--\n"
	"body { background: white }\n"
	"h1 { text-align: center; color: red }\n"
	"-->\n"
	"</style>\n"
	"<title>%d %s</title>\n"
	"</head>\n"
	"<body lang=\"ja\">\n"
	"<h1 class=\"center\">%d %s</h1>\n"
	"<p>\n",
  	getResponseCode(thd->rsCode), getResponsePhrase(thd->rsCode),
	getResponseCode(thd->rsCode), getResponsePhrase(thd->rsCode));

  /*
   * send body part
   */
  switch (thd->rsCode) {
    case CODE_BAD_REQUEST:			/* 400 */
      iosPuts(thd, MSG_BAD_REQUEST);
      break;
    case CODE_UNAUTHORIZED:			/* 401 */
      iosPrintf(thd, MSG_UNAUTHORIZED, thd->requestURI);
      break;
    case CODE_FORBIDDEN:			/* 403 */
      iosPrintf(thd, MSG_FORBIDDEN, thd->requestURI);
      break;
    case CODE_NOT_FOUND:			/* 404 */
      iosPrintf(thd, MSG_NOT_FOUND, thd->requestURI);
      break;
    case CODE_METHOD_NOT_ALLOWED:		/* 405 */
      iosPrintf(thd, MSG_METHOD_NOT_ALLOWED, thd->method.name);
      break;
    case CODE_NOT_ACCEPTABLE:			/* 406 */
      iosPrintf(thd, MSG_NOT_ACCEPTABLE, thd->mtype);
      break;
    case CODE_REQUEST_TIMEOUT:			/* 408 */
      iosPuts(thd, MSG_REQUEST_TIMEOUT);
      break;
    case CODE_PRECONDITION_FAILED:		/* 412 */
      iosPuts(thd, MSG_PRECONDITION_FAILED);
      break;
    case CODE_REQUEST_URI_TOO_LONG:		/* 414 */
      iosPuts(thd, MSG_REQUEST_URI_TOO_LONG);
      break;
    case CODE_INTERNAL_SERVER_ERROR:		/* 500 */
      iosPrintf(thd, MSG_INTERNAL_SERVER_ERROR, thd->requestURI);
      break;
    case CODE_NOT_IMPLEMENTED:			/* 501 */
      iosPrintf(thd, MSG_NOT_IMPLEMENTED, thd->requestURI);
      break;
    case CODE_HTTP_VERSION_NOT_SUPPORTED:	/* 505 */
      iosPuts(thd, MSG_HTTP_VERSION_NOT_SUPPORTED);
      break;
    default:
      break;
  }

  /*
   * send footer part
   */
  iosPrintf(thd,
	"</p>\n"
	"<hr>\n"
	"<address>\n"
	"Powered by&nbsp;"
	"<a href=\"%s\">%s</a>\n"
	"&nbsp;&lt;<a href=\"mailto:%s\">%s</a>&gt;\n"
	"</address>\n"
	"</body>\n"
	"</html>\n",
	SERVER_ADDRESS, SERVER_VERSION,
	httpd->vhost->serverAdmin, httpd->vhost->serverAdmin);
  iosFlush(thd);
}

/*
 * Function Name:	sendCustomizedErrorMessage
 * Description:		send a customized error message
 * Arguments:		thread
 * Return Value:	OK: TRUE, otherwise: FALSE
 */
static bool_t
sendCustomizedErrorMessage(thd)
	register Thread *	thd;
{
  String	path;	/* customized error document path */

  path = getErrorDocument(getResponseCode(thd->rsCode));
  if (path == NULL) {
    return FALSE;
  }

  XFree(thd->path);
  if (xasprintf(&thd->path, "%s%s", httpd->vhost->docRoot, path) == -1) {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    return FALSE;
  }

  thd->fp = fopen(thd->path, "r");
  if (thd->fp == NULL) {
    return FALSE;
  }

  sendingEntity(thd);

  return TRUE;
}

