/*****************************************************************************
 *
 * FILE:	io.c
 * DESCRIPTION:	input/output functions
 * DATE:	Sun, Oct  3 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: io.c,v 1.27.2.2 2002/01/27 16:09:19 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif	/* __STDC__ */
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	oneRead(Thread *, String);
static int	fillReadBuffer(Thread *);
static int	flushWriteBuffer(Thread *);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	nread
 * Description:		read data from buffer
 * Arguments:		thread, buffer, buffer size
 * Return Value:	success: n, EOF: 0, error: -1
 */
int
nread(thd, buf, bufSize)
	register Thread *	thd;
	String			buf;
	size_t			bufSize;
{
  int	ret;
  int	n;

  if (thd->ios.update == TRUE) {
    flushWriteBuffer(thd);
  }

  ret = 0;
  if (thd->ios.next >= thd->ios.total) {
    ret = fillReadBuffer(thd);
    if (ret != 1) {
      return ret;	/* error */
    }
  }

  for (n = 0; thd->ios.next < thd->ios.total && n < bufSize; ) {
    buf[n++] = thd->ios.buf[thd->ios.next++];
  }

  return n;
}

/*
 * Function Name:	readline
 * Description:		read a line into buffer
 * Arguments:		thread, buffer, buffer size
 * Return Value:	Successful: n, EOF: 0, error: -1
 */
int
readline(thd, buf, bufSize)
	register Thread *	thd;
	String			buf;
	size_t			bufSize;
{
  int		n;
  String	ptr;

  ptr = buf;
  for (n = 0; n < bufSize - 1; n++) {
    char	c;
    int		ret = oneRead(thd, &c);
    if (ret == 1) {
      if (n < bufSize - 1) {
	*ptr++ = c;
      }
      if (c == LF) {
	break;
      }
    }
    else if (ret == 0) {
      if (n == 0) {
	return 0;	/* EOF */
      }
    }
    else {
      return -1;	/* error */
    }
  }
  *ptr = EOL;

  return n;
}


/*
 * Function Name:	oneRead
 * Description:		read one character from buffer
 * Arguments:		thread, buffer
 * Return Value:	success: 1, error: -1
 */
static int
oneRead(thd, buf)
	register Thread *	thd;
	String			buf;
{
  int	ret;

  if (thd->ios.update == TRUE) {
    flushWriteBuffer(thd);
  }

  ret = 1;
  if (thd->ios.next >= thd->ios.total) {
    ret = fillReadBuffer(thd);
    if (ret != 1) {
      return ret;	/* error */
    }
  }
  *buf = thd->ios.buf[thd->ios.next++];

  return ret;
}


/*
 * Function Name:	fillReadBuffer
 * Description:		fill the request buffer with read data
 * Arguments:		thread
 * Return Value:	success: 1, EOF: 0, error: -1
 */
static int
fillReadBuffer(thd)
	register Thread *	thd;
{
  time_t	limit;

  time(&limit);
again:
#if	USE_SSL
  if (thd->SSL.active == TRUE && SSL_get_fd(thd->SSL.ssl) ==  thd->sd) {
    thd->ios.total = SSL_read(thd->SSL.ssl, thd->ios.buf, sizeof(thd->ios.buf));
  }
  else {
    thd->ios.total = read(thd->sd, thd->ios.buf, sizeof(thd->ios.buf));
  }
#else
  thd->ios.total = read(thd->sd, thd->ios.buf, sizeof(thd->ios.buf));
#endif	/* USE_SSL */
  switch (thd->ios.total) {
    case -1:
      if (errno == EINTR) {
	goto again;
      }
      else if (errno == EAGAIN || errno == EWOULDBLOCK) {
	usleep(IO_USLEEP_TIME);
	if (thd->keepAlive.state && (time(NULL) - limit > 1)) {
	  ;	/* nothing */
	}
	else if (time(NULL) - limit > httpd->timeout) {
	  setResponseCode(thd, CODE_REQUEST_TIMEOUT);
	}
	else {
	  goto again;
	}
      }
      else {
	setResponseCode(thd, CODE_BAD_REQUEST);
      }
      return -1;
    case 0:
      return 0;
    default:
      thd->ios.next = 0;
      break;
  }

  return 1;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	iosFlush
 * Description:		force data in buf to output
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
int
iosFlush(thd)
	register Thread *	thd;
{
  return flushWriteBuffer(thd);
}

/*
 * Function Name:	iosWrite
 * Description:		data in buf to output
 * Arguments:		thread, buffer, buffer size
 * Return Value:	successful: outputed bytes, otherwise: -1
 */
int
iosWrite(thd, buf, size)
	register Thread *	thd;
	const char *		buf;
	size_t			size;
{
  int	n;

  for (n = 0; ; ) {
    while (thd->ios.next < sizeof(thd->ios.buf) && n < size) {
      thd->ios.buf[thd->ios.next++] = buf[n++];
    }
    if (thd->ios.next >= sizeof(thd->ios.buf)) {
      if (flushWriteBuffer(thd) == -1) {
	return -1;
      }
    }
    if (n >= size) {
      break;
    }
  }
  thd->ios.update = TRUE;

  return n;
}

/*
 * Function Name:	iosPuts
 * Description:		output string
 * Arguments:		thread, string
 * Return Value:	successful: 0, otherwise: -1
 */
int
iosPuts(thd, str)
	register Thread *	thd;
	String			str;
{
  size_t	len;

  for (len = 0; ; ) {
    while (thd->ios.next < sizeof(thd->ios.buf) && str[len] != EOL) {
      thd->ios.buf[thd->ios.next++] = str[len++];
    }
    if (thd->ios.next >= sizeof(thd->ios.buf)) {
      if (flushWriteBuffer(thd) == -1) {
	return -1;
      }
    }
    if (str[len] == EOL) {
      break;
    }
  }
  thd->ios.update = TRUE;

  return 0;
}


/*
 * Function Name:	iosPrintf
 * Description:		output with format
 * Arguments:		thread, format, arguments
 * Return Value:	successful: sent bytes, otherwise: -1
 */
#define	LONGINT		0x001	/* long integer */
#define	HEXDIGIT	0x002	/* hex-digit */
int
#if	__STDC__
iosPrintf(register Thread *thd, const char *fmt, ...)
#else
iosPrintf(thd, fmt, va_alist)
	register Thread *	thd;
	char *			fmt;
	va_dcl
#endif	/* __STDC__ */
{
  char		ch;
  int		flags;
  int		d;	/* for int type */
  long int	l;	/* for long int type */
  String	s;	/* for char * type */
  size_t	len;
  int		n;
  va_list	ap;
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */

  for (;;) {
    for ( ; (ch = *fmt) != EOL && ch != '%'; fmt++) {
      if (thd->ios.next < sizeof(thd->ios.buf)) {
	thd->ios.buf[thd->ios.next++] = ch;
      }
      if (thd->ios.next >= sizeof(thd->ios.buf)) {
	flushWriteBuffer(thd);
      }
    }
    thd->ios.update = TRUE;

    if (ch == EOL) {
      goto done;
    }
    fmt++;	/* skip over '%' */
    flags = 0;
    len	  = 0;
restart:
    ch = *fmt++;
    switch (ch) {
      case ' ':
	goto restart;
      case '\'':
      case '#': case '*': case '-': case '+': case '.':
      case '0':
      case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	goto restart;
      case 'c':
	/* Note: char is promoted to int. */
	if (thd->ios.next >= sizeof(thd->ios.buf)) {
	  flushWriteBuffer(thd);
	}
	thd->ios.buf[thd->ios.next++] = va_arg(ap, int);
	continue;
      case 's':
	s	= va_arg(ap, char *);
	len	= strlen(s);
	break;
      case 'l':
	flags |= LONGINT;
	goto restart;
      case 'x':
	flags |= HEXDIGIT;
      case 'd':
      case 'i':
	if (flags & LONGINT) {
	  l = va_arg(ap, long int);
	}
	else {
	  d = va_arg(ap, int);
	}
	if (flags & HEXDIGIT) {
	  if (flags & LONGINT) {
	    s = itoxa((unsigned long int)l);
	  }
	  else {
	    s = itoxa((unsigned int)d);
	  }
	}
	else {
	  if (flags & LONGINT) {
	    s = itoa(l);
	  }
	  else {
	    s = itoa(d);
	  }
	}
	len = strlen(s);
	break;
      default:
	if (ch == EOL) {
	  goto done;
	}
	goto restart;
    }

    for (n = 0; ; ) {
      while (thd->ios.next < sizeof(thd->ios.buf) && n < len) {
	thd->ios.buf[thd->ios.next++] = s[n++];
      }
      if (thd->ios.next >= sizeof(thd->ios.buf)) {
	flushWriteBuffer(thd);
      }
      if (s[n] == EOL) {
	break;
      }
    }
    thd->ios.update = TRUE;
  }

done:
  va_end(ap);

  return 0;
}


/*
 * Function Name:	flushWriteBuffer
 * Description:		flush out data in buffer
 * Arguments:		thread
 * Return Value:	successful: sent bytes, otherwise: -1
 */
static int
flushWriteBuffer(thd)
	register Thread *	thd;
{
  int	n;
  int	total;

  n = 0;
  if (thd->ios.update == FALSE) {
    return n;
  }
#if	DEBUG
  fprintf(stderr, "[io]: next(%d) total(%d) bufsize(%d)\n",
	  thd->ios.next, thd->ios.total, sizeof(thd->ios.buf));
#endif	/* DEBUG */

  total = thd->ios.total > 0 ? thd->ios.total : 0;
#if	USE_SSL
  if (thd->SSL.active == TRUE && SSL_get_fd(thd->SSL.ssl) == thd->sd) {
    while (total < thd->ios.next) {
ssl_again:
      n = SSL_write(thd->SSL.ssl, &thd->ios.buf[total], (thd->ios.next - total));
      if (n == -1) {
	if (errno == EAGAIN) {
	  usleep(IO_USLEEP_TIME);
	  goto ssl_again;
	}
	break;	/* error, exit while() block */
      }
      total += n;
    }
  }
  else {
    while (total < thd->ios.next) {
again:
      n = write(thd->sd, &thd->ios.buf[total], (size_t)(thd->ios.next - total));
      if (n == -1) {
	if (errno == EAGAIN) {
	  usleep(IO_USLEEP_TIME);
	  goto again;
	}
	break;	/* error, exit while() block */
      }
      total += n;
    }
  }
#else	/* NOUSE_SSL */
  while (total < thd->ios.next) {
again:
    n = write(thd->sd, &thd->ios.buf[total], (size_t)(thd->ios.next - total));
    if (n == -1) {
      if (errno == EAGAIN) {
	usleep(IO_USLEEP_TIME);
	goto again;
      }
      break;	/* error, exit while() block */
    }
    total += n;
  }
#endif	/* USE_SSL */
  thd->ios.update = FALSE;
  thd->ios.total  = 0;
  thd->ios.next	  = 0;
  memset(thd->ios.buf, 0, sizeof(thd->ios.buf));

  /* client close the connection! */
  if (errno == EPIPE) {
    thd->state = TERMINATED;
  }

  return n;
}

