/*****************************************************************************
 *
 * FILE:	mime.c
 * DESCRIPTION:	register/get MIME types
 * DATE:	Sat, Sep 18 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mime.c,v 1.22.2.3 2002/01/30 13:26:33 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <stdlib.h>
#include <string.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of local macros
 *
 *****************************************************************************/
#define	MTYPE_HTML	"html"
#define	MTYPE_GIF	"gif"
#define	MTYPE_JPEG	"jpeg"
#define	MTYPE_PNG	"png"
#define	MTYPE_GZIP	"gz"
#define	MTYPE_BZIP	"bz2"
#define	MTYPE_ZIP	"zip"
#define	MTYPE_LHA	"lha"
#define	MTYPE_SSI	"shtml"


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static Hash *	mimeTable[MIME_TABLE_SIZE];


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static String		getMimeType(String);
static String		quickCheck(Thread *);
static unsigned int	hash(String);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	initMimeTypeTable
 * Description:		initialize hash table
 * Arguments:		none
 * Return Value:	none
 */
void
initMimeTypeTable(void)
{
  register int	i;

  for (i = 0; i < MIME_TABLE_SIZE; i++) {
    mimeTable[i] = NULL;
  }
}

/*
 * Function Name:	freeMimeTypeTable
 * Description:		free the memory area of mime type
 * Arguments:		none
 * Return Value:	none
 */
void
freeMimeTypeTable(void)
{
  register int	i;

  for (i = 0; i < MIME_TABLE_SIZE; i++) {
    register Hash *	hp;
    register Hash *	tmp;

    if (mimeTable[i] == NULL) {
      continue;
    }
    for (hp = mimeTable[i]; hp != NULL; hp = tmp) {
      tmp = hp->next;
      XFree((String)hp->key);
      XFree((String)hp->value);
      XFree((Hash *)hp);
    }
    mimeTable[i] = NULL;
  }
}


/*
 * Function Name:	registerMimeType
 * Description:		Registers the suffix and the mimetype to hash table.
 * Arguments:		suffix, mimetype
 * Return Value:	none
 */
void
registerMimeType(suffix, type)
	String	suffix;
	String	type;
{
  int		h;	/* for hash value	*/
  Hash *	hp;	/* for current pointer	*/

  if (suffix == NULL) {
    return;
  }

  if (getMimeType(suffix) != NULL) {
    return;	/* already registered */
  }

  hp = (Hash *)xcalloc(1, sizeof(Hash));
  if (hp == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }

  h		= hash(suffix);
  hp->key	= xstrdup(suffix);
  hp->value	= xstrdup(type);
  hp->next	= mimeTable[h];
  mimeTable[h]	= hp;
}

/*
 * Function Name:	getMimeType
 * Description:		get mime-type
 * Arguments:		suffix
 * Return Value:	mime-type
 */
static String
getMimeType(suffix)
	String	suffix;
{
  register Hash *	hp;

  for (hp = mimeTable[hash(suffix)]; hp != NULL; hp = hp->next) {
    if (strcmp(hp->key, suffix) == 0) {
      return hp->value;
    }
  }

  return NULL;
}


/*
 * Function Name:	setMimeType
 * Description:		set mimetype of supplied file.
 * Arguments:		thread
 * Return Value:	none
 */
void
setMimeType(thd)
	register Thread *	thd;
{
  String	suffix;	/* suffix of file */

  suffix = quickCheck(thd);
  if (suffix == NULL) {
    suffix = strrchr(thd->path, '.');
    if (suffix++ == NULL) {	/* no suffix */
      thd->mtype = httpd->vhost->defaultType;
      return;
    }
  }

  thd->mtype = getMimeType(suffix);
  if (thd->mtype == NULL) {
    thd->mtype = httpd->vhost->defaultType;
  }
  else if (strcmp(thd->mtype, CGI_MIME_TYPE) == 0) {	/* CGI */
    thd->ctype = C_CGI;
    thd->state = CGI_READY;
  }
  else if (strcmp(thd->mtype, SSI_MIME_TYPE) == 0) {	/* SSI */
    thd->ctype = C_SSI;
    thd->state = SSI_READY;
  }
}

/*
 * Function Name:	quickCheck
 * Description:		check the file type
 * Arguments:		thread
 * Return Value:	mimetype
 */
#define	CHECK_SIZE	32
static String
quickCheck(thd)
	Thread *	thd;
{
  char	buf[CHECK_SIZE];

  /* not yet opened */
  if (thd->fp == NULL) {
    return NULL;
  }

  memset(buf, 0, sizeof(buf));
  if (fread(buf, sizeof(char), sizeof(buf), thd->fp) <= 0) {
    return NULL;
  }
  (void)fseek(thd->fp, 0L, SEEK_SET);

  if (memcmp(buf, "<!DOCTYPE HTML", 14) == 0 ||
      memcmp(buf, "<!doctype html", 14) == 0 ||
      memcmp(buf, "<HEAD",	     5) == 0 ||
      memcmp(buf, "<head",	     5) == 0 ||
      memcmp(buf, "<TITLE",	     6) == 0 ||
      memcmp(buf, "<title",	     6) == 0 ||
      memcmp(buf, "<HTML",	     5) == 0 ||
      memcmp(buf, "<html",	     5) == 0 ||
      memcmp(buf, "<BODY",	     5) == 0 ||
      memcmp(buf, "<body",	     5) == 0) {
    String p = strstr(thd->path, ".shtml");
    if (p != NULL && *(p + 6) == EOL) {
      return MTYPE_SSI;
    }
    return MTYPE_HTML;
  }
  else if (memcmp(buf, "GIF8", 4) == 0) {
    return MTYPE_GIF;
  }
  else if (memcmp(buf+6, "JFIF", 4) == 0) {
    return MTYPE_JPEG;
  }
  else if (memcmp(buf+1, "PNG", 3) == 0) {
    return MTYPE_PNG;
  }
  else if (memcmp(buf, "\037\213", 2) == 0) {
    return MTYPE_GZIP;
  }
  else if (memcmp(buf, "BZ", 2) == 0) {
    return MTYPE_BZIP;
  }
  else if (memcmp(buf, "PK", 2) == 0) {
    return MTYPE_ZIP;
  }
  else if (memcmp(buf+2, "-lh", 3) == 0 || memcmp(buf+2, "-lz", 3) == 0) {
    return MTYPE_LHA;
  }

  return NULL;	/* no assumable */
}

/*
 * Function Name:	hash
 * Description:		calculates a hash value of string
 * Arguments:		string
 * Return Value:	hash value
 */
static unsigned int
hash(str)
	String	str;
{
  unsigned int	h;

  for (h = 0; *str != EOL; str++) {
    h = (67 * h + *str) % MIME_TABLE_SIZE;
  }

  return h;
}

