/*****************************************************************************
 *
 * FILE:	dir.c
 * DESCRIPTION:	library for directory listing module of wyvern
 * DATE:	Wed, May  2 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2001-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: dir.c,v 1.18.2.5 2002/03/07 10:46:22 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <stdlib.h>
#include <string.h>
#if	HAVE_DIRENT_H
#include <dirent.h>
#else
#define	dirent	direct
#if	HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif	/* HAVE_SYS_NDIR_H */
#if	HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif	/* HAVE_SYS_DIR_H */
#if	HAVE_NDIR_H
#include <ndir.h>
#endif	/* HAVE_NDIR_H */
#endif	/* HAVE_DIRENT_H */
#ifdef HAVE_D_NAMLEN
#define NAMLEN(d)	(d->d_namlen)
#else	/* !HAVE_D_NAMLEN */
#define NAMLEN(d)	(strlen(d->d_name))
#endif	/* !HAVE_D_NAMLEN */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "wyvern.h"
#include "functions.h"
#include "dir.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
extern void	directoryListing(Thread *);


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
typedef struct _DirIndex {	/* directory cache structure */
  String		path;	/* path of data file */
  String		name;	/* path of cache file */
  time_t		mtime;	/* last modified time */
  struct _DirIndex *	next;
  struct _DirIndex *	prev;
} DirIndex;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void		openDirectoryCache(Thread *, DirIndex *);
static void		updateDirectoryCache(Thread *, DirIndex *);
static void		quickSort(String *, int, int);
static void		swap(String *, int, int);
static void		initDirCache(void);
static void		insertDirCache(DirIndex *, DirIndex *);
static DirIndex *	removeDirCache(DirIndex *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static DirIndex	dirCache[DIR_TABLE_SIZE];	/* cache entry */
static DirIndex	dirList;
static void	(*makeContentOfHTML)() = NULL;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	initLibDir
 * Description:		initialize directory library
 * Arguments:		function's pointer (for makeContentOfHTML())
 * Return Value:	none
 */
void
initLibDir(func)
	void	(*func)();
{
  makeContentOfHTML = func;
  initDirCache();
}

/*
 * Function Name:	freeLibDir
 * Description:		unlink directory cache file
 * Arguments:		none
 * Return Value:	none
 */
void
freeLibDir(void)
{
  register int	i;

  for (i = 0; i < DIR_TABLE_SIZE; i++) {
    XFree(dirCache[i].path);
    (void)unlink(dirCache[i].name);
    XFree(dirCache[i].name);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	directoryListing
 * Description:		file listing in the directory
 * Arguments:		thread
 * Return Value:	none
 * Note:		this function is called by dlsym().
 */
void
directoryListing(thd)
	register Thread *	thd;
{
  register DirIndex *	dp;

  /* need directory listing? */
  if (thd->ctype != C_DIR) {
    return;
  }
 
  /* no allowed directory walking */
  if (thd->method.type == M_HEAD) {
    setResponseCode(thd, CODE_FORBIDDEN);
#if	DEBUG
    fprintf(stderr, "[dir.c] method is head = %s\n", thd->path);
#endif	/* DEBUG */
    return;
  }

  /* search cache (dirList) */
  for (dp = dirList.next; dp != &dirList; dp = dp->next) {
    if (dp->mtime == 0) {
      continue;
    }
    if (strcmp(thd->path, dp->path) == 0) {	/* cache hit */
      if (dp->mtime == thd->file.st_mtime) {
#if	DEBUG
	fputs("[dir.c]: cache hit!\n", stderr);
#endif	/* DEBUG */
	openDirectoryCache(thd, dp);
      }
      else {	/* directory was modified. */
#if	DEBUG
	fputs("[dir.c]: cache update!\n", stderr);
#endif	/* DEBUG */
	updateDirectoryCache(thd, dp);
      }
      return;
    }
  }

  /* get top entry from list, and move one to end of list */
  dp = removeDirCache(dirList.next);
  insertDirCache(&dirList, dp);

  /* register new entry */
  XFree(dp->path);
  dp->path = xstrdup(thd->path);

  XFree(dp->name);
  if (xasprintf(&dp->name,  "%s/dir_cache%lx%lx%lx",
		httpd->cacheDir, (long)thd->file.st_ino,
		(long)thd->file.st_size, thd->file.st_mtime) == -1) {
    dp->mtime = 0;	/* mark of no use */
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    return;
  }
  dp->mtime = thd->file.st_mtime;

  updateDirectoryCache(thd, dp);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	openDirectoryCache
 * Description:		open cache file
 * Arguments:		thread, cache
 * Return Value:	none
 */
static void
openDirectoryCache(thd, cache)
	register Thread *	thd;
	DirIndex *		cache;
{
  struct stat	sbuf;

  /* close previous opened file pointer */
  if (thd->fp != NULL) {
    XFClose(thd->fp);
  }

  /* open directory listed file */
  thd->fp = fopen(cache->name, "r");
  if (thd->fp == NULL) {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    return;
  }

  /* set actual file status */
  if (fstat(fileno(thd->fp), &sbuf) == 0) {
    thd->file = sbuf;
  }

  /* set mime-type */
  setMimeType(thd);

  /* set entity tag */
  setETag(thd);

  setResponseCode(thd, CODE_REQUEST_OK);
}

/*
 * Function Name:	updateDirectoryCache
 * Description:		update cache file for directory indexing
 * Arguments:		thread, cache
 * Return Value:	none
 */
static void
updateDirectoryCache(thd, cache)
	register Thread *	thd;
	DirIndex *		cache;
{
  register DIR *	dfd;
  struct dirent *	dp;
  struct stat		sbuf;
  size_t		items;	/* total entries in directory */
  String *		dir;
  int			dnum = 0;
  FILE *		fout;
  int			i;

  /*
   * directory walking
   */
  dfd = opendir(thd->path);
  if (dfd == NULL) {
    setResponseCode(thd, CODE_FORBIDDEN);
#if	DEBUG
    fprintf(stderr, "[dir.c] failed to open: %s\n", thd->path);
#endif	/* DEBUG */
    return;
  }

  /*
   * estimate total entries in directory
   */
#ifdef HAVE_DIRFD
  if (fstat(dirfd(dfd), &sbuf) == -1) {
#else
  if (fstat(dfd->dd_fd, &sbuf) == -1) {
#endif	/* !HAVE_DIRFD */
    closedir(dfd);
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    return;
  }
  items = (sbuf.st_size / 24);
  dir	= (String *)xcalloc(items, sizeof(String));
  if (dir == NULL) {
    closedir(dfd);
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    wvSyslog(LOG_ALERT, ERR_OUT_OF_MEMORY);
    return;
  }

  /*
   * directory scanning ...
   */
  while ((dp = readdir(dfd)) != NULL) {
    String	p;

    /* skip dot files */
    if (dp->d_name[0] == '.' || dnum > MAX_DIR_ENTRY - 1) {
      continue;
    }

    /* skip CVS/RCS directory */
    if (strncmp(dp->d_name, "CVS", 3) == 0 ||
	strncmp(dp->d_name, "RCS", 3) == 0) {
      continue;
    }
    /* skip *,? and *~ files (for CVS and emacs series) */
    if (dp->d_name[NAMLEN(dp) - 2] == ',' ||
	dp->d_name[NAMLEN(dp) - 1] == '~') {
      continue;
    }
    /* skip #* and ,* files (for MH) */
    if (dp->d_name[0] == '#' || dp->d_name[0] == ',') {
      continue;
    }
    /* skip *.{bak,org} files (for vim series) */
    p = strrchr(dp->d_name, '.');
    if (p != NULL) {
      if (strcasecmp(p, ".bak") == 0 || strcasecmp(p, ".org") == 0) {
	continue;
      }
    }

    dir[dnum] = xstrdup(dp->d_name);
    if (dir[dnum] == NULL) {
      closedir(dfd);
      setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
      wvSyslog(LOG_ALERT, ERR_OUT_OF_MEMORY);
      goto error;
    }
    dnum++;
  }
  closedir(dfd);

  /*
   * sort directory indexing
   */
  quickSort(dir, 0, dnum - 1);

  /*
   * update cache file
   */
  fout = fopen(cache->name, "w");
  if (fout == NULL) {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    goto error;
  }
  (*makeContentOfHTML)(thd, fout, dir, dnum);
  fclose(fout);

  /* bind new file descriptor to current thread */
  openDirectoryCache(thd, cache);

error:
  for (i = 0; i < dnum; i++) {
    xfree((String)dir[i]);
  }
  xfree((String *)dir);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	quickSort
 * Description:		sort string by upper order
 * Arguments:		string array, left value, right value
 * Return Value:	none
 */
static void
quickSort(s, left, right)
	String	s[];
	int	left, right;
{
  int	i, last;

  if (left >= right) {
    return;
  }

  swap(s, left, (left + right) / 2);
  last = left;

  for (i = left + 1; i <= right; i++) {
    if (strcmp(s[i], s[left]) < 0) {
      swap(s, ++last, i);
    }
  }

  swap(s, left, last);
  quickSort(s, left, last - 1);
  quickSort(s, last + 1, right);
}

/*
 * Function Name:	swap
 * Description:		swap the string
 * Arguments:		string array, swapped position (1), swapped position (2)
 * Return Value:	none
 */
static void
swap(s, i, j)
	String	s[];
	int	i, j;
{
  String	t;

  t	= s[i];
  s[i]	= s[j];
  s[j]	= t;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	initDirCache
 * Description:		initialize dirCache called at only once
 * Arguments:		none
 * Return Value:	none
 */
static void
initDirCache(void)
{
  int	i;

  dirList.next = dirList.prev = &dirList;
  for (i = 0; i < DIR_TABLE_SIZE; i++) {
    dirCache[i].path = NULL;
    dirCache[i].name = NULL;
    dirCache[i].mtime = 0;
    insertDirCache(&dirList, &dirCache[i]);
  }
}

/*
 * Function Name:	insertDirCache
 * Description:		insert 'new' before 'pos'
 * Arguments:		position, new slot
 * Return Value:	none
 */
static void
insertDirCache(pos, new)
	register DirIndex *	pos;
	register DirIndex *	new;
{
  pos->prev->next = new;
  new->prev	  = pos->prev;
  new->next	  = pos;
  pos->prev	  = new;
}

/*
 * Function Name:	removeDirCache
 * Description:		remove the slot from dirCache
 * Arguments:		removed slot
 * Return Value:	removed slot
 */
static DirIndex *
removeDirCache(x)
	register DirIndex *	x;
{
  x->prev->next = x->next;
  x->next->prev = x->prev;
  return x;
}

