/*****************************************************************************
 *
 * FILE:	mode_access.c
 * DESCRIPTION:	access control module for Wyvern
 * DATE:	Sat, Sep 16 2000
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 2000-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (c) 2000-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_access.c,v 1.21.2.2 2004/12/19 13:52:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
extern bool_t	allowHost(register Thread *);


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	SPECIAL_KEYWORD	"DEFAULT"
typedef struct _Restrict {
  String		host;
  bool_t		allow;
  struct _Restrict *	next;
} Restrict;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static Restrict *	accessHostTable[ACCESS_TABLE_SIZE];
static bool_t		empty = TRUE;	/* above table empty ? TRUE : FALSE */
static bool_t		defaultPermission;	/* default access permission */


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void		registerHost(FILE *);
static String		checkHost(String);
static bool_t		compareAddr(String, String);
static unsigned int	hash(String);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	_init
 * Description:		initialize the host access list (automatic loading)
 * Arguments:		none
 * Return Value:	none
 */
void
_init(void)
{
  int		i;
  FILE *	fin;

#if	DEBUG
  fputs("[mod_access]: plugged-in\n", stderr);
#endif	/* DEBUG */

  /* initialize host table */
  for (i = 0; i < ACCESS_TABLE_SIZE; i++) {
    accessHostTable[i] = NULL;
  }

  /* register allowed access hosts */
  fin = fopen(GLOBAL_HOSTS_ALLOW, "r");
  if (fin != NULL) {
    defaultPermission = TRUE;	/* allow all hosts to access */
    registerHost(fin);
    fclose(fin);
  }
}

/*
 * Function Name:	_fini
 * Description:		release the memory area of access lists
 *			(automatic loading)
 * Arguments:		none
 * Return Value:	none
 */
void
_fini(void)
{
  int	i;

  for (i = 0; i < ACCESS_TABLE_SIZE; i++) {
    register Restrict *	rp;
    register Restrict *	tmp;

    if (accessHostTable[i] == NULL) {
      continue;
    }
    for (rp = accessHostTable[i]; rp != NULL; rp = tmp) {
      tmp = rp->next;
      xfree((Restrict *)rp);
    }
    accessHostTable[i] = NULL;
  }
  empty = TRUE;
}


/*
 * Function Name:	allowHost
 * Description:		permit the access to the host
 * Arguments:		thread
 * Return Value:	allowed host: TRUE, denied host: FALSE
 */
bool_t
allowHost(thd)
	register Thread *	thd;
{
  struct hostent *	hp;
  register Restrict *	rp;

  /* no table entry: allow all hosts */
  if (empty) {
    return TRUE;
  }

  /* get client hostname with DNS */
  XFree(thd->client.name);
#if	USE_IPV6
  hp = gethostbyaddr((String)&thd->client.addr.sin6_addr,
		      sizeof(thd->client.addr.sin6_addr),
		      thd->client.addr.sin6_family);
#else
  hp = gethostbyaddr((String)&thd->client.addr.sin_addr,
		      sizeof(thd->client.addr.sin_addr), PF_INET);
#endif	/* USE_IPV6 */
  if (hp != NULL) {
    thd->client.name = xstrdup(hp->h_name);
  }

  /* check the client hostname */
  for (rp = accessHostTable[hash(thd->client.name)]; rp != NULL;
       rp = rp->next) {
    if (compareAddr(thd->client.name, rp->host) == TRUE) {
      return rp->allow;
    }
  }

  /* check the client IP address */
  for (rp = accessHostTable[hash(thd->client.ip)]; rp != NULL; rp = rp->next) {
    if (compareAddr(thd->client.ip, rp->host) == TRUE) {
      return rp->allow;
    }
  }

  return defaultPermission;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	registerHost
 * Description:		register host to the host access list
 * Arguments:		FILE pointer
 * Return Value:	none
 */
static void
registerHost(fin)
	register FILE *	fin;
{
#define	MAX_ARGS	2
#define	A_NAME		0	/* hostname */
#define	A_PERMISSION	1	/* permission */
  char		buf[MAX_HOSTNAME_LEN];

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    String *		ap;
    String		args[MAX_ARGS];
    String		p = buf;
    bool_t		perm;
    Restrict *		rp;
    unsigned int	h;

    if (skipLine(buf)) {
      continue;
    }

    for (ap = args; (*ap = strsep(&p, " \t\r\n")) != NULL; ) {
      if (**ap != EOL) {
	if (++ap >= &args[MAX_ARGS]) {
	  break;
	}
      }
    }

    if (strcasecmp(args[A_PERMISSION], "allow") == 0) {
      perm = TRUE;
    }
    else if (strcasecmp(args[A_PERMISSION], "deny") == 0) {
      perm = FALSE;
    }

    if (strcasecmp(SPECIAL_KEYWORD, args[A_NAME]) == 0) {
#if	DEBUG
      fprintf(stderr, "[mod_access]: default access = %s\n", perm ? "allow" : "deny");
#endif	/* DEBUG */
      defaultPermission = perm;
      goto next;
    }

    if (checkHost(args[A_NAME]) != NULL) {
      goto next;	/* already registered */
    }

    rp = (Restrict *)xcalloc(1, sizeof(Restrict));
    if (rp == NULL) {
      die(ERR_OUT_OF_MEMORY);
    }

    h			= hash(args[A_NAME]);
    rp->host		= xstrdup(args[A_NAME]);
    rp->allow		= perm;
    rp->next		= accessHostTable[h];
    accessHostTable[h]	= rp;

next:
    memset(buf, 0, sizeof(buf));
    empty = FALSE;
  } /* while end */
}

/*
 * Function Name:	checkHost
 * Description:		check host in the host access list
 * Arguments:		host
 * Return Value:	host or NULL
 */
static String
checkHost(host)
	String	host;
{
  register Restrict *	rp;

  for (rp = accessHostTable[hash(host)]; rp != NULL; rp = rp->next) {
    if (strcmp(rp->host, host) == 0) {
      return rp->host;
    }
  }

  return NULL;
}

/*
 * Function Name:	compareAddr
 * Description:		check the hostname, IP addr and the access list
 * Arguments:		Hostname/IP addr, compared one
 * Return Value:	Matching: TRUE, Otherwise: FALSE
 */
static bool_t
compareAddr(name, pat)
	String	name;
	String	pat;
{
  int	nlen = strlen(name);
  int	plen = strlen(pat);

  /*
   * check to match the name with the pattern like as '.ad.jp'
   */
  if (pat[0] == '.') {
    if (nlen >= plen) {
      int	offset = nlen - plen;
      if (strncmp(name + offset, pat, (size_t)plen) == 0) {
	return TRUE;
      }
    }
  }
  /*
   * check to match the name with the pattern like as '133.24.' or IP addr
   */
  else if (pat[0] >= '0' && pat[0] <= '9') {
    if (strncmp(name, pat, (size_t)plen) == 0) {
      return TRUE;	/* match! */
    }
  }
  /*
   * check to match the name with the pattern of FQDN 
   */
  else {
    if (strcmp(name, pat) == 0) {
      return TRUE;	/* match! */
    }
  }

  return FALSE;	/* no match */
}


/*
 * Function Name:	hash
 * Description:		calculates the hash value of string
 * Arguments:		string
 * Return Value:	hash value
 */
static unsigned int
hash(str)
	String	str;
{
  register unsigned char *	p;
  unsigned int			h;

  /* ivalid pointer */
  if (str == NULL) {
    return 0;
  }

  p = (unsigned char *)str;
  h = *p;
  do {
    h += *p++;
  } while (*p != '.' && *p != EOL);

  return h % ACCESS_TABLE_SIZE;
}

