/*****************************************************************************
 *
 * FILE:	mod_ssi.c
 * DESCRIPTION:	SSI module for wyvern
 * DATE:	Sat, Sep 16 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2000-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_ssi.c,v 1.36.2.6 2002/02/05 12:16:41 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_VFORK_H
#include <vfork.h>
#endif	/* HAVE_VFORK_H */
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <time.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
extern void	parseSSI(Thread *);


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#ifndef	MAP_FAILED
#define	MAP_FAILED	((caddr_t)-1)
#endif	/* MAP_FAILED */
#define	U_KILO		(1024L)
#define	U_MEGA		(1024L*1024L)
#define	U_GIGA		(1024L*1024L*1024L)
#define	TIMEFMT_LEN	64

typedef enum {
  SSI_CONFIG,
  SSI_INCLUDE,
  SSI_ECHO,
  SSI_FSIZE,
  SSI_FLASTMOD,
  SSI_EXEC
} ssi_directives_t;

typedef enum {	/* ssi token parsing state */
  ST_GND,
  ST_LBR,
  ST_EXC,
  ST_HYP1,
  ST_HYP2,
  ST_CMD
} state_t;

typedef enum {
  SZ_BYTES,
  SZ_ABBREV
} sizefmt_t;

typedef enum {
  ET_TAG,
  ET_VALUE,
  ET_FORMAT
} error_ssi_t;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static FILE *	parse(Thread *);
static void	parseDirectiveOfSSI(Thread *, FILE *, String);

static void	do_config(Thread *, String, String);
static void	do_include(Thread *, FILE *, String, String);
static void	do_echo(Thread *, FILE *, String, String);
static void	do_fsize(Thread *, FILE *, String, String);
static void	do_flastmod(Thread *, FILE *, String, String);
static void	do_exec(Thread *, FILE *, String, String);
static void	runProgram(Thread *, FILE *, const String);

static void	getFormattedTime(Thread *, FILE *, time_t, bool_t);
static String	getFileName(Thread *, String, String);

static void	errorSSI(error_ssi_t, String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static char		timefmt[TIMEFMT_LEN];
static sizefmt_t	sizefmt;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	_init
 * DESCRIPTION:		register SHTML mime-type (automatic loading)
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
_init(void)
{
#if	DEBUG
  fputs("[mod_ssi]: plugged-in\n", stderr);
#endif	/* DEBUG */
  registerMimeType("shtml", SSI_MIME_TYPE);
}

/*
 * FUNCTION NAME:	parseSSI
 * DESCRIPTION:		address an SSI content
 * ARGUMENTS:		thread
 * RETURN VALUE:	none
 */
void
parseSSI(thd)
	register Thread *	thd;
{
  FILE *	fin;
  struct stat	sbuf;

  /* default formats */
  memset(timefmt, 0, sizeof(timefmt));
  				/* ex. Sat Nov 27 01:26:37 JST 1999	*/
  strcpy(timefmt, "%a %b %e %T %Z %Y");	/* never overflow */
  sizefmt = SZ_BYTES;

  /* parsing */
  fin = parse(thd);
  if (fin == NULL ||
      fstat(fileno(fin), &sbuf) == -1)  {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    return;
  }

  /* reopen parsed SSI file */
  thd->fp	= fin;
  thd->state	= SEND_READY;
  thd->ctype	= C_STD;
  thd->mtype	= TYPE_TEXT_HTML;
  thd->file	= sbuf;	/* set actual content length */
  setETag(thd);
}

/*
 * FUNCTION NAME:	parse
 * DESCRIPTION:		parsing HTML file
 * ARGUMENTS:		thread
 * RETURN VALUE:	successful: FILE pointer, otherwise: NULL
 */
static FILE *
parse(thd)
	register Thread *	thd;
{
#define	SSI_BUFSIZE	256
  String		ssiFile;
  register FILE *	fin;
  register FILE *	fout;
  int			ch;
  state_t		st	= ST_GND;
  bool_t		flush	= FALSE;
  char			directive[SSI_BUFSIZE];
  String		p	= directive;
  char			text[SSI_BUFSIZE];
  String		q	= text;

  /* open SSI file */
  fin = fopen(thd->path, "r");
  if (fin == NULL) {
    return NULL;
  }

  /* open temporary file */
  if (xasprintf(&ssiFile, "%s/ssifile%u",
		httpd->tmpDir, (unsigned int)thd) == -1) {
    return NULL;
  }
  fout = fopen(ssiFile, "w");
  if (fout == NULL) {
    fclose(fin); fin = NULL;
    goto done;
  }

  memset(directive, 0, sizeof(directive));
  memset(text, 0, sizeof(text));
  while ((ch = fgetc(fin)) != EOF) {
    switch (st) {
      case ST_LBR:
	*q++ = ch;
	if (ch == '!')	{ st = ST_EXC; }
	else		{ st = ST_GND; flush = TRUE; }
	break;
      case ST_EXC:
	*q++ = ch;
	if (ch == '-')	{ st = ST_HYP1; }
	else		{ st = ST_GND; flush = TRUE; }
	break;
      case ST_HYP1:
	*q++ = ch;
	if (ch == '-')	{ st = ST_HYP2; }
	else		{ st = ST_GND; flush = TRUE; }
	break;
      case ST_HYP2:
	*q++ = ch;
	if (ch == '#')	{ st = ST_CMD; }
	else		{ st = ST_GND; flush = TRUE; }
	break;
      case ST_CMD:
	if (ch == '-') {
	  int	c0 = fgetc(fin);
	  int	c1 = fgetc(fin);

	  if (c0 == '-' && c1 == '>') {
	    *p = EOL;
	    parseDirectiveOfSSI(thd, fout, directive);
	    memset(directive, 0, sizeof(directive));
	    p = directive;
	    memset(text, 0, sizeof(text));
	    q  = text;
	    st = ST_GND;
	  }
	  else {
	    *p++ = ch;
	    *p++ = c0;
	    *p++ = c1;
	  }
	}
	else {
	  *p++ = ch;
	}
	break;
      case ST_GND:
      default:
	if (ch == '<') {
	  st = ST_LBR;
	  *q++ = ch;
	}
	else {
	  fputc(ch, fout);
	}
	break;
    }
    if (flush) {
      fputs(text, fout);
      memset(text, 0, sizeof(text));
      q	    = text;
      flush = FALSE;
    }
  }
  fclose(fout);
  fclose(fin);

  /* reopen parsed SSI file */
  fin = fopen(ssiFile, "r");
  if (fin == NULL) {
    goto done;
  }
  /* the file will be removed when it's closed. */
  (void)unlink(ssiFile);

done:
  xfree((String)ssiFile);
  return fin;
}

/*
 * FUNCTION NAME:	parseDirectiveOfSSI
 * DESCRIPTION:		parsing SSI directive
 * ARGUMENTS:		thread, mapped buffer
 * RETURN VALUE:	none
 */
static void
parseDirectiveOfSSI(thd, fout, cmd)
	Thread *	thd;
	register FILE *	fout;
	String		cmd;
{
  String	directive = NULL;
  String	tag	  = NULL;
  String	val	  = NULL;
  String	p	  = NULL;

  /* SSI directive */
  directive  = cmd;
  directive += strspn(cmd, " \t\r\n");

  /* SSI tag */
  tag	     = directive;
  tag	     = (String)strpbrk(directive, " \t\r\n");
  if (tag == NULL) {
    wvSyslog(LOG_ERR, "illegal SSI format: %s\n", directive);
    return;
  }
  *tag++     = EOL;
  tag	    += strspn(tag, " \t\r\n=");

  /* SSI value */
  val	     = tag;
  val	     = (String)strpbrk(tag, " \t\r\n=");
  if (val == NULL) {
    wvSyslog(LOG_ERR, "SSI %s needs a tag\n", directive);
    return;
  }
  *val++     = EOL;
  val	     = (String)strchr(val, '"');
  if (val++ == NULL) {
    wvSyslog(LOG_ERR, "SSI %s's %s needs a value\n", directive, tag);
    return;
  }
  p = (String)strchr(val, '"');
  if (p == NULL) {
    wvSyslog(LOG_ERR, "SSI %s's value of %s is not closed by \".\n",
	     directive, tag);
    return;
  }
  *p = EOL;

  /* SSI directive */
  if (strncmp(directive, "config", 6) == 0) {
    do_config(thd, tag, val);
  }
  else if (strncmp(directive, "include", 7) == 0) {
    do_include(thd, fout, tag, val);
  }
  else if (strncmp(directive, "echo", 4) == 0) {
    do_echo(thd, fout, tag, val);
  }
  else if (strncmp(directive, "fsize", 5) == 0) {
    do_fsize(thd, fout, tag, val);
  }
  else if (strncmp(directive, "flastmod", 8) == 0) {
    do_flastmod(thd, fout, tag, val);
  }
  else if (strncmp(directive, "exec", 4) == 0) {
    do_exec(thd, fout, tag, val);
  }
  else {
    errorSSI(ET_FORMAT, directive);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	do_config
 * DESCRIPTION:		address SSI config format
 * ARGUMENTS:		thread, tag, val
 * RETURN VALUE:	none
 */
static void
do_config(thd, tag, val)
	Thread *	thd;
	String		tag;
	String		val;
{
  if (strncmp(tag, "timefmt", 7) == 0) {
    strncpy(timefmt, val, TIMEFMT_LEN);
  }
  else if (strncmp(tag, "sizefmt", 7) == 0) {
    if (strncmp(val, "bytes", 5) == 0) {
      sizefmt = SZ_BYTES;
    }
    else if (strncmp(val, "abbrev", 6) == 0) {
      sizefmt = SZ_ABBREV;
    }
    else {
      errorSSI(ET_VALUE, val);
    }
  }
  else {
    errorSSI(ET_TAG, tag);
  }
}

/*
 * FUNCTION NAME:	do_include
 * DESCRIPTION:		address SSI include format
 * ARGUMENTS:		thread, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_include(thd, fout, tag, val)
	Thread *	thd;
	register FILE *	fout;
	String		tag;
	String		val;
{
  register FILE *	fin;
  char			buf[STD_BUFSIZE];
  String		fpath = getFileName(thd, tag, val);

  if (fpath == NULL) {
    wvSyslog(LOG_ERR, "SSI failed to get file path: %s=%s\n", tag, val);
    return;
  }

  fin = fopen(fpath, "r");
  if (fin == NULL) {
    return;
  }
  while (!feof(fin)) {
    size_t	size = fread(buf, sizeof(char), (size_t)sizeof(buf), fin);
    fwrite(buf, sizeof(char), size, fout);
  }
  fclose(fin);
}

/*
 * FUNCTION NAME:	do_echo
 * DESCRIPTION:		address SSI echo format
 * ARGUMENTS:		thread, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_echo(thd, fout, tag, val)
	Thread *	thd;
	register FILE *	fout;
	String		tag;
	String		val;
{
  time_t	cur_time;

  if (strncmp(tag, "var", 3) != 0) {
    errorSSI(ET_TAG, tag);
    return;
  }

  time(&cur_time);	/* set current time in cur_time */
  if (strcmp(val, "DOCUMENT_NAME") == 0) {
    /* the current filename */
    String	p = (String)strrchr(thd->path, '/');
    if (p != NULL) {
      fputs(p + 1, fout);
    }
  }
  else if (strcmp(val, "DOCUMENT_URI") == 0) {
    /* the virtual path to this document (such as /docs/tutorials/foo.shtml) */
    fputs(thd->requestURI, fout);
  }
  else if (strcmp(val, "QUERY_STRING_UNESCAPED") == 0) {
    if (thd->cgi.query_string != NULL) {
      fputs(thd->cgi.query_string, fout);
    }
  }
  else if (strcmp(val, "DATE_LOCAL") == 0) {
    /* the current date, local time zone */
    getFormattedTime(thd, fout, cur_time, FALSE);	/* localtime */
  }
  else if (strcmp(val, "DATE_GMT") == 0) {
    /* the current date in Greenwich mean  time */
    getFormattedTime(thd, fout, cur_time, TRUE);	/* Greenwich */
  }
  else if (strcmp(val, "LAST_MODIFIED") == 0) {
    struct stat	sbuf;

    if (lstat(thd->path, &sbuf) == 0) {
      /* the last modification date of the current document */
      getFormattedTime(thd, fout, sbuf.st_mtime, FALSE);	/* localtime */
    }
  }
}

/*
 * FUNCTION NAME:	do_fsize
 * DESCRIPTION:		address SSI fsize format
 * ARGUMENTS:		thread, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_fsize(thd, fout, tag, val)
	Thread *	thd;
	register FILE *	fout;
	String		tag;
	String		val;
{
  struct stat	sbuf;
  String	fpath = getFileName(thd, tag, val);

  if (fpath == NULL) {
    wvSyslog(LOG_ERR, "SSI failed to get file path: %s=%s\n", tag, val);
    return;
  }

  if (lstat(fpath, &sbuf) == -1) {
    wvSyslog(LOG_ERR, "SSI failed to get file status: %s\n", fpath);
    return;
  }

  switch (sizefmt) {
    case SZ_BYTES:
      fprintf(fout, "%ld", (long)sbuf.st_size);
      break;
    case SZ_ABBREV:
      if (sbuf.st_size < 1024) {
	fprintf(fout, "%ld", (long)sbuf.st_size);
      }
      else if (sbuf.st_size < 1024 * 1024) {
	fprintf(fout, "%ld.%ldK", (long)sbuf.st_size / U_KILO,
				  ((long)sbuf.st_size % U_KILO) * 10 / U_KILO);
      }
      else if (sbuf.st_size < 1024 * 1024 * 1024) {
	fprintf(fout, "%ld.%ldM", (long)sbuf.st_size / U_MEGA,
				  ((long)sbuf.st_size % U_MEGA) * 10 / U_MEGA);
      }
      else {
	fprintf(fout, "%ld.%ldG", (long)sbuf.st_size / U_GIGA,
				  ((long)sbuf.st_size % U_GIGA) * 10 / U_GIGA);
      }
    default:
      break;
  }
}

/*
 * FUNCTION NAME:	do_flastmod
 * DESCRIPTION:		address SSI flastmod format
 * ARGUMENTS:		thread, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_flastmod(thd, fout, tag, val)
	Thread *	thd;
	register FILE *	fout;
	String		tag;
	String		val;
{
  struct stat	sbuf;
  String	fpath = getFileName(thd, tag, val);

  if (fpath == NULL) {
    wvSyslog(LOG_ERR, "SSI failed to get file path: %s=%s\n", tag, val);
    return;
  }

  if (lstat(fpath, &sbuf) == -1) {
    wvSyslog(LOG_ERR, "SSI failed to get file status: %s\n", fpath);
    return;
  }

  getFormattedTime(thd, fout, sbuf.st_mtime, FALSE);	/* localtime */
}

/*
 * FUNCTION NAME:	do_exec
 * DESCRIPTION:		address SSI exec format
 * ARGUMENTS:		thread, output file pointer, tag, val
 * RETURN VALUE:	none
 */
static void
do_exec(thd, fout, tag, val)
	Thread *	thd;
	register FILE *	fout;
	String		tag;
	String		val;
{
  if (strncmp(tag, "cmd", 3) == 0) {
    runProgram(thd, fout, val);
  }
#if	0
  else if (strncmp(tag, "cgi", 3) == 0) {
    /* not implemented... */
  }
#endif
  else {
    errorSSI(ET_TAG, tag);
  }
}

/*
 * FUNCTION NAME:	runProgram
 * DESCRIPTION:		execute a program for SSI's cmd
 * ARGUMENTS:		thread, output file pointer, command
 * RETURN VALUE:	none
 */
#define	READ_FD		0	/* descriptor to read */
#define	WRITE_FD	1	/* descriptor to write */
#define	MAX_ARGS	5	/* maximum number of arguments for program */
static void
runProgram(thd, fout, cmd)
	Thread *	thd;
	register FILE *	fout;
	const String	cmd;
{
  char		cwd[MAX_PATH_LEN];	/* command working directory */
  String	args[MAX_ARGS];
  String *	ap;
  String	str = cmd;
  int		c;
  int		pfd[2];
  pid_t		pid;
  FILE *	fin;
  char		buf[STD_BUFSIZE];
  /*
   * shell meta characters
   * \134 = backslash
   * \042 = double quotation
   * \012 = new line
   * \015 = carriage return
   */
  const char	sh_meta[] = "&;`'\134\042|*?~<>^()[]{}$\012\015";

  /* Does cmd include shell meta characters? */
  if (strpbrk(cmd, sh_meta) != NULL) {
    wvSyslog(LOG_EMERG,
	     "unsecure/untrusted command string [%s] for #cmd in SSI\n", cmd);
    return;
  }

  /* parsing arguments */
  for (c = 0, ap = args; (*ap = strsep(&str, " \t")) != NULL; c++) {
    args[c] = *ap;
    if (**ap != EOL) {
      if (++ap >= &args[MAX_ARGS]) {
	wvSyslog(LOG_ERR,
		 "unexpected too many arguments [%s] for #cmd in SSI\n", cmd);
	return;	/* error: arguments overflow */
      }
    }
  }
  args[c] = EOL;

  /* set working directory */
  memset(cwd, 0, sizeof(cwd));
  if (xstrncpy(cwd, thd->path, sizeof(cwd)) != NULL &&
      setScriptEnvironment(thd) != FALSE &&	/* create an environments */
      pipe(pfd) != -1) {/* make pipeline between parent process and child.*/
    String	p = strrchr(cwd, '/');
    if (p != NULL) {
      *p = EOL;
    }
  }
  else {
    return;
  }

  /*
   * execute a program
   */
  pid = vfork();
  switch (pid) {
    case -1:	/* fork error */
      wvSyslog(LOG_ERR, "failed to fork process for [%s] #cmd in SSI\n", cmd);
      break;

    case 0:	/* child process */
      xclose(pfd[READ_FD]);
      dup2(pfd[WRITE_FD], STDOUT_FILENO);
      dup2(pfd[WRITE_FD], STDERR_FILENO);
      if (chdir(cwd) == -1 || execve(args[0], args, thd->cgi.env) == -1) {
	wvSyslog(LOG_ERR, "failed to exec [%s] #cmd in SSI\n", args[0]);
      }
      _exit(1);

    default:	/* parent process */
      xclose(pfd[WRITE_FD]);
      fin = fdopen(pfd[READ_FD], "r");
      if (fin != NULL) {
	while (!feof(fin)) {
	  size_t size = fread(buf, sizeof(char), (size_t)sizeof(buf), fin);
	  fwrite(buf, sizeof(char), size, fout);
	}
	fclose(fin);
      }
      break;
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	getFormattedTime
 * DESCRIPTION:		get the formatted time
 * ARGUMENTS:		thread, output file pointer,
 *			time, gmt flag (TRUE: GMT, FALSE: Localtime)
 * RETURN VALUE:	none
 */
static void
getFormattedTime(thd, fout, t, gmt)
	Thread *	thd;
	register FILE *	fout;
	time_t		t;
	bool_t		gmt;
{
  struct tm *	tp;
  char		tbuf[MIDDLE_BUFSIZE];

  tp = gmt ? gmtime(&t) : localtime(&t);
  memset(tbuf, 0, sizeof(tbuf));
  if (strftime(tbuf, sizeof(tbuf), timefmt, tp) > 0) {
    fputs(tbuf, fout);
  }
}

/*
 * FUNCTION NAME:	getFileName
 * DESCRIPTION:		get filename
 * ARGUMENTS:		thread, tag, val
 * RETURN VALUE:	filename
 */
static String
getFileName(thd, tag, val)
	Thread *	thd;
	String		tag;
	String		val;
{
  static String	paths[] = {"/etc/", "/bin/", "/sbin/", "/usr/", "/var/"};
  static String	fpath	= NULL;

  if (fpath != NULL) {
    XFree(fpath);
  }

  if (strncmp(tag, "file", 4) == 0) {
    int	i;

    for (i = 0; i < sizeof(paths)/sizeof(String); i++) {
      if (strncmp(val, paths[i], strlen(paths[i])) == 0) {
	wvSyslog(LOG_EMERG, "invalid file access [%s] in SSI\n", val);
	return NULL;
      }
    }
    if (xasprintf(&fpath, "%s", val) == -1) {
      wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
      return NULL;
    }
  }
  else if (strncmp(tag, "virtual", 7) == 0) {
    int	n;

    if (thd->userDir != NULL) {
      n = xasprintf(&fpath, "%s/%s", thd->userDir, val);
    }
    else {
      n = xasprintf(&fpath, "%s/%s", httpd->vhost->docRoot, val);
    }
    if (n == -1) {
      wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    }
  }
  else {
    errorSSI(ET_TAG, tag);
  }

  return fpath;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	errorSSI
 * DESCRIPTION:		no support SSI tag/value/format
 * ARGUMENTS:		type, name
 * RETURN VALUE:	none
 */
static void
errorSSI(type, name)
	error_ssi_t	type;
	String		name;
{
  switch (type) {
    case ET_TAG:
      wvSyslog(LOG_WARNING, "no support SSI tag: %s\n", name);
      break;
    case ET_VALUE:
      wvSyslog(LOG_WARNING, "no support SSI value: %s\n", name);
      break;
    case ET_FORMAT:
      wvSyslog(LOG_WARNING, "no support SSI format: #%s\n", name);
    default:
      break;
  }
}

