/*****************************************************************************
 *
 * FILE:	request.c
 * DESCRIPTION:	address the request URI from client
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: request.c,v 1.86.2.9 2002/02/13 05:02:03 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <sys/wait.h>
#if	TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if	TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif	/* TM_IN_SYS_TIME */
#endif	/* TIME_WITH_SYS_TIME */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	SOCKET_ERROR	(-1)
typedef enum {
  SOCKET_NORMAL,
  SOCKET_SECURE
} socket_t;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static int	maxfd;
static fd_set	rfds, rset;
static fd_set	wfds, wset;
static long	interval = DEF_TIMEOUT;	/* the renewal timeout */


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	initMask(void);
static void	newRequest(socket_t);
#if	USE_IPV6
static int	acceptRequest(int, struct sockaddr_storage *);
#else
static int	acceptRequest(int, struct sockaddr_in *);
#endif	/* USE_IPV6 */
static bool_t	accessControl(Thread *);
static void	requestOk(void);
static bool_t	timeout(Thread *);
static int	checkTimeOut(void);
static bool_t	recvRequestLine(Thread *);
static bool_t	recvHeader(Thread *);
static void	recvEntity(Thread *);
static String	getAcceptField(Thread *, String);
static void	checkAcceptField(Thread *);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	initMask
 * Description:		initialize mask of file descriptors
 * Arguments:		none
 * Return Value:	none
 */
static void
initMask(void)
{
  /* set socket mask */
  FD_ZERO(&rfds);
  FD_ZERO(&wfds);
  FD_SET(httpd->socket.soc, &rfds);
#if	USE_SSL
  FD_SET(httpd->socket.ssl, &rfds);
  maxfd = MAX(httpd->socket.soc, httpd->socket.ssl);
#else
  maxfd = httpd->socket.soc;
#endif	/* USE_SSL */
}

/*
 * Function Name:	foreverRunning
 * Description:		address the request from each client
 * Arguments:		none
 * Return Value:	none
 */
void
foreverRunning(void)
{
  struct timeval	expired;
  int			ret;	/* return value of select() */

  /*
   * request loop
   */
  initMask();
  for (;;) {
    /* set timeout */
    expired.tv_sec  = interval;
    expired.tv_usec = 0;

    /* clear mask */
    rset = rfds;
    wset = wfds;
    ret  = select(maxfd + 1, &rset, &wset, NULL, &expired);
    switch (ret) {
      /* TIMEOUT */
      case 0:
	if (checkTimeOut() == 0) {	/* no expired by timeout */
	  interval = interval < MAX_TIMEOUT ? interval * 8 : MAX_TIMEOUT;
	}
	sweepThreadQueue();
#if	DEBUG
	fprintf(stderr, "[select]: timeout; waiting %d sec\n", (int)interval);
#endif	/* DEBUG */
	if (httpd->reboot) {		/* reboot */
	  Reconfiguration();		/* read wyvern.conf again */
	  initMask();
	  interval = DEF_TIMEOUT;	/* reset timeout */
	}
        break;

      /* ERROR */
      case -1:
	if (errno == EINTR) {
	  if (httpd->reboot) {
	    interval = 1;	/* set timeout for quick response */
	  }
	  continue;
	}
	die("failed to select in request loop ... it's critical.\n");

      /* READY */
      default:
	interval = DEF_TIMEOUT;		/* reset timeout */
	if (FD_ISSET(httpd->socket.soc, &rset)) {
	  newRequest(SOCKET_NORMAL);
	}
#if	USE_SSL
	else if (FD_ISSET(httpd->socket.ssl, &rset)) {
	  newRequest(SOCKET_SECURE);
	}
#endif	/* USE_SSL */
	else {	/* The request is already in the thread queue. */
	  requestOk();
	}
    } /* switch end */
  } /* for end */
}

/*
 * Function Name:	newRequest
 * Description:		receive new request and perform it
 * Arguments:		socket_t
 * Return Value:	none
 */
static void
newRequest(stype)
	socket_t	stype;
{
#if	USE_IPV6
  struct sockaddr_storage	cliaddr;
#else
  struct sockaddr_in	cliaddr;
#endif	/* USE_IPV6 */
  int		sd;	/* new socket descriptor */
  Thread *	thd;

  /* new request from client */
  switch (stype) {
    case SOCKET_NORMAL:
      sd = acceptRequest(httpd->socket.soc, &cliaddr);
      break;
#if	USE_SSL
    case SOCKET_SECURE:
      sd  = acceptRequest(httpd->socket.ssl, &cliaddr);
      break;
#endif	/* USE_SSL */
    default:
      return;
  }
  if (sd == SOCKET_ERROR) {
    return;
  }

#if	DEBUG
  fputs("[accept]: new connection\n", stderr);
#endif	/* DEBUG */
  thd = newThread(sd);/* yield new thread */
  if (thd == NULL) {	/* thread queue is full */
    xclose(sd);
    wvSyslog(LOG_ERR, "*** THREAD QUEUE OVERFLOWED ***\n");
#if	DEBUG
    fputs("[accept]: Overflow connection\n", stderr);
#endif	/* DEBUG */
    return;
  }

  /* set client hostname, IP address and remote port */
  setClientName(thd, &cliaddr);
#if	DEBUG
  fprintf(stderr, "[client]: port = %d\n", thd->client.port);
#endif	/* DEBUG */
#if	USE_SSL
  switch (stype) {
    case SOCKET_SECURE:
      switch (sslAcceptNewConnection(thd)) {
	case SSL_ERROR:
	  detachThread(thd);
	  wvSyslog(LOG_ERR, "failed to setup SSL connection\n");
#if	DEBUG
	  fprintf(stderr, "[SSL(%d)]: failed...\n", thd->client.port);
#endif	/* DEBUG */
	  return;
	case SSL_OK:
	  if (accessControl(thd) == FALSE) {
	    detachThread(thd);
	    return;
	  }
	  break;
	case SSL_RETRY:
	default:
	  break;
      }
      break;
    default:
      if (accessControl(thd) == FALSE) {
	detachThread(thd);
	return;
      }
      break;
  }
#else
  if (accessControl(thd) == FALSE) {
    detachThread(thd);
    return;
  }
#endif	/* USE_SSL */

  /* update the max file descriptor */
  FD_SET(thd->sd, &rfds);
  rset = rfds;	/* for quick response */
  if (thd->sd > maxfd) {
    maxfd = thd->sd;
  }

  requestOk();
}

/*
 * Function Name:	acceptRequest
 * Description:		accept request from client and return new descriptor
 * Arguments:		socket, struct sockaddr * (for client address)
 * Return Value:	new socket descriptor, or -1
 */
static int
acceptRequest(soc, ret)
	int			soc;
#if	USE_IPV6
	struct sockaddr_storage	*	ret;
#else
	struct sockaddr_in *	ret;
#endif	/* USE_IPV6 */
{
#if	USE_IPV6
  struct sockaddr_storage	cliaddr;
#else
  struct sockaddr_in	cliaddr;
#endif	/* USE_IPV6 */
  int	clilen;
  int	sd;	/* new socket descriptor */
  int	on;	/* flag to set socket option */
  int	flags;

  /* new request from client via SSL */
  clilen = sizeof(cliaddr);
  sd	 = accept(soc, (struct sockaddr *)&cliaddr, (socklen_t *)&clilen);
  if (sd == -1) {
    return SOCKET_ERROR;
  }
  *ret = cliaddr;	/* set client address */

  /* allow reuse of local addresses ( error is ignored ) */
  on = 1;
  setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));

  /* change new socket descriptor to non-blocking mode */
  flags = fcntl(sd, F_GETFL, 0);
  if (flags != -1) {
    flags |= O_NONBLOCK;
    if (fcntl(sd, F_SETFL, flags) == -1) {
      wvSyslog(LOG_ERR, "failed to change non-blocking mode\n");
      xclose(sd);
      return SOCKET_ERROR;
    }
  }

  return sd;
}


/*
 * Function Name:	accessControl
 * Description:		check access control list
 * Arguments:		thread
 * Return Value:	Access OK: TRUE, otherwise: FALSE
 */
static bool_t
accessControl(thd)
	register Thread *	thd;
{
  /* identification remote user by RFC1413 */
  if (httpd->mod.ident.set != NULL) {
    httpd->mod.ident.set(thd);
  }

  /* check the host level access control */
  if (httpd->mod.access.check != NULL &&
      httpd->mod.access.check(thd) == FALSE) {
#if	DEBUG
    fputs("[access]: denied...\n", stderr);
#endif	/* DEBUG */
    wvSyslog(LOG_NOTICE, "refused a connection from %s\n",
	     thd->client.name == NULL ? thd->client.ip : thd->client.name);
    return FALSE;
  }

  return TRUE;
}


/*
 * Function Name:	requestOk
 * Description:		perform action for request
 * Arguments:		none
 * Return Value:	none
 */
static void
requestOk(void)
{
  register Thread *	thd;

  for (thd = getThreadHead(); thd != NULL; thd = thd->next) {
    timeout(thd);	/* check timeout */
    switch (thd->state) {
      case RECV_READY:
	if (FD_ISSET(thd->sd, &rset)) {
#if	DEBUG
	  fprintf(stderr, "[receive(%d)]: request from client\n",
		  thd->client.port);
#endif	/* DEBUG */
	  if (recvRequestLine(thd) == TRUE) {
	    if (recvHeader(thd) == TRUE) {
	      allowRequest(thd);
	      thd->atime = time(NULL);	/* update */
	    }
	    FD_SET(thd->sd, &wfds);	/* reponse immediately */
	    wset = wfds;		/* for quick response */
	  }
	  else if (thd->state != IN_ERROR &&
		   thd->keepAlive.state != KA_CONTINUE) {
	    FD_CLR(thd->sd, &rfds);
#if	DEBUG
	    fprintf(stderr, "[receive(%d)]: keep-alive state = %d\n",
		    thd->client.port, thd->keepAlive.state);
#endif	/* DEBUG */
	  }
	  else {	/* Error */
	    FD_SET(thd->sd, &wfds);	/* reponse immediately */
	    wset = wfds;		/* for quick response */
#if	DEBUG
	    fprintf(stderr, "[receive(%d)]: ka_max=%d, state=%d, code=%d\n",
		    thd->client.port, thd->keepAlive.max, thd->state,
		    thd->rsCode);
#endif	/* DEBUG */
	  }
	}
	break;
      case IN_ERROR:
	thd->keepAlive.state = KA_DISABLED;	/* disable Keep-Alive mode */
#if	DEBUG
	  fprintf(stderr, "[error(%d)]: response to client\n", thd->client.port);
#endif	/* DEBUG */
      case EMB_READY:
      case SEND_READY:
	if (FD_ISSET(thd->sd, &wset)) {
#if	DEBUG
	  fprintf(stderr, "[send(%d)]: response to client\n", thd->client.port);
#endif	/* DEBUG */
	  sendResponse(thd);	/* send response */
	}
	break;
      case NOW_SENDING:
	if (FD_ISSET(thd->sd, &wset)) {
#if	DEBUG
	  fprintf(stderr, "[send(%d)]: now sending...\n", thd->client.port);
#endif	/* DEBUG */
	  sendingEntity(thd);
	  thd->atime = time(NULL);	/* update */
	}
	break;
      case TERMINATED:
#if	DEBUG
	  fprintf(stderr, "[terminated(%d)]: #%d ", thd->client.port, thd->no);
#endif	/* DEBUG */
	cleanupThread(thd);
	if (thd->keepAlive.state == KA_CONTINUE && timeout(thd) == FALSE) {
	  thd->atime	 = time(NULL);	/* update */
	  thd->state	 = RECV_READY;
	  interval	 = thd->keepAlive.timeout;
#if	DEBUG
	  fputs("connection continued\n\n", stderr);
#endif	/* DEBUG */
	}
	else {
	  FD_CLR(thd->sd, &rfds);
	  FD_CLR(thd->sd, &wfds);
	  detachThread(thd);
#if	DEBUG
	  fputs("thread terminated...\n\n", stderr);
#endif	/* DEBUG */
	}
	break;
      case CGI_READY:
	if (httpd->mod.cgi.run != NULL) {
	  httpd->mod.cgi.run(thd);
	  FD_SET(fileno(thd->fp), &rfds);
	  rset = rfds;	/* for quick response */
	}
	else {
	  setResponseCode(thd, CODE_BAD_REQUEST);
	}
	break;
      case CGI_WAITING:
	httpd->mod.cgi.parseHeader(thd);
	FD_CLR(fileno(thd->fp), &rfds);
	break;
      case SSI_READY:
	if (httpd->mod.ssi.run != NULL) {
	  httpd->mod.ssi.run(thd);
	}
	else {
	  setResponseCode(thd, CODE_BAD_REQUEST);
	}
	break;
#if	USE_SSL
      case SSL_READY:
	switch (sslEstablishConnection(thd)) {
	  case SSL_ERROR:
	    FD_CLR(thd->sd, &rfds);
	    detachThread(thd);
	    wvSyslog(LOG_ERR, "failed to setup SSL connection\n");
#if	DEBUG
	    fprintf(stderr, "[SSL<%d>]: failed...\n", thd->client.port);
#endif	/* DEBUG */
	    break;
	  case SSL_RETRY:
	    if (accessControl(thd) == FALSE) {
	      FD_CLR(thd->sd, &rfds);
	      detachThread(thd);
	    }
#if	DEBUG
	    fprintf(stderr, "DEBUG[SSL<%d>]: Try again...\n", thd->client.port);
#endif	/* DEBUG */
	    break;
	  case SSL_OK:
	  default:
#if	DEBUG
	    fprintf(stderr, "[SSL<%d>]: Ready -> OK\n", thd->client.port);
#endif	/* DEBUG */
	    break;
	}
	break;
#endif	/* USE_SSL */
      default:
#if	DEBUG
	fprintf(stderr, "[detached(%d)]: state=%d\n", thd->no, thd->state);
#endif	/* DEBUG */
	break;
    }
  }
}

/*
 * Function Name:	timeout
 * Description:		check timeout of thread
 * Arguments:		thread
 * Return Value:	timeouted: TRUE, otherwise: FALSE
 */
static bool_t
timeout(thd)
	register Thread *	thd;
{
  time_t	now;
  time_t	passed;

  if (thd->state == IN_ERROR ||		/* already an error happened */
      thd->state == TERMINATED) {	/* already done */
    return FALSE;
  }

  /* set current time */
  time(&now);

  /* calculate passed time */
  passed = now - thd->atime;

  /* check timeout for Keep-Alive connection */
  if (thd->keepAlive.state && passed > thd->keepAlive.timeout) { /* timeout */
      thd->state	   = TERMINATED;
      thd->keepAlive.state = KA_CLOSE;
#if	DEBUG
      fprintf(stderr, "[timeout(%d)]: Keep-Alive is timeout: %s\n",
	      thd->client.port, thd->requestURI);
#endif	/* DEBUG */
  }
  else if (passed > httpd->timeout) {
#if	DEBUG
    fprintf(stderr, "[timeout(%d)]: request timeout: %s\n",
	    thd->client.port, thd->requestURI);
#endif	/* DEBUG */
    setResponseCode(thd, CODE_REQUEST_TIMEOUT);	/* timeout */
  }
  else {
    return FALSE;	/* continued */
  }

  return TRUE;		/* timeout */
}

/*
 * Function Name:	checkTimeOut
 * Description:		check socket's timeout of each client
 * Arguments:		thread
 * Return Value:	number of timeouted thread
 */
static int
checkTimeOut(void)
{
  register Thread *	thd;	/* thread */
  register int		count;	/* number of timeouted thread */

  for (count = 0, thd = getThreadHead(); thd != NULL; thd = thd->next) {
    if (timeout(thd)) {
      count++;
    }
  }

  return count;
}


/*
 * Function Name:	allowRequest
 * Description:		check the request from a client
 * Arguments:		thread
 * Return Value:	none
 */
void
allowRequest(thd)
	register Thread *	thd;
{
  /* switch virtual host */
  if (switchVirtualHost(thd) == FALSE) {
    return;
  }

  /* parse request URI */
  if (parseURI(thd) == FALSE) {
    return;
  }

  /* check the user side access control */
  if (httpd->mod.auth.allow != NULL && httpd->mod.auth.allow(thd) == FALSE) {
    return;	/* don't allow user's access */
  }

  /* address by method */
  switch (thd->method.type) {
    case M_HEAD:
    case M_GET:
      openRequestFile(thd);
      /* Do directory listing */
      if (thd->ctype == C_DIR) {
	if (httpd->mod.dir.listing != NULL) {
	  httpd->mod.dir.listing(thd);
	}
	else {
	  setResponseCode(thd, CODE_FORBIDDEN);
#if	DEBUG
	  fprintf(stderr, "[Dir] no load module = %s\n", thd->path);
#endif	/* DEBUG */
	}
      }
      checkAcceptField(thd);
      break;
    case M_POST:
      recvEntity(thd);
      break;
#if	0	/* Not implemented */
    case M_PUT:
    case M_DELETE:
    case M_MOVE:
#if	0
      setResponseCode(thd, CODE_METHOD_NOT_ALLOWED);
      break;
#endif
    case M_OPTIONS:
    case M_TRACE:
    case M_CONNECT:
    case M_SEARCH:
#endif
    default:
      setResponseCode(thd, CODE_NOT_IMPLEMENTED);
      return;
  }
}

/*
 * Function Name:	recvRequestLine
 * Description:		receive the request line from a client
 * Arguments:		thread
 * Return Value:	error is FALSE, otherwise TRUE
 */
static bool_t
recvRequestLine(thd)
	register Thread *	thd;
{
  int		n;
  char		buf[BUF_SIZE];
  String *	ap;
  String	args[3];
  String	p;

  /* skip first white lines (for old buggy browser) */
  memset(buf, 0, sizeof(buf));
  while ((n = readline(thd, buf, (size_t)sizeof(buf))) > 0) {
    if (buf[0] != CR && buf[1] != LF) {	/* check CRLF */
      break;
    }
  }
#if	DEBUG
  fprintf(stderr, "[readline]: read characters = #%d (%d)\n", n, thd->state);
#endif	/* DEBUG */

  /* check error */
  if (thd->state == IN_ERROR) {
    return FALSE;
  }
  else if (n < 0) {
    if (thd->keepAlive.state == KA_CONTINUE && thd->state == RECV_READY) {
#if	DEBUG
      fprintf(stderr, "[keep-alive(%d)] n=%d ka_max=%d\n",
	      thd->client.port, n, thd->keepAlive.max);
#endif	/* DEBUG */
      return FALSE;
    }
    setResponseCode(thd, CODE_BAD_REQUEST);
    return FALSE;
  }
  else if (n >= sizeof(thd->requestURI)) {
    setResponseCode(thd, CODE_REQUEST_URI_TOO_LONG);
    /* set request URI to record error_log */
    (void)xstrncpy(thd->requestURI, buf, sizeof(thd->requestURI));
    return FALSE;
  }

  /*
   * divide the request line into the method, URI and protocol
   */
  for (n = 0, p = chomp(buf), ap = args; (*ap = strsep(&p, " ")) != NULL; ) {
    if (**ap != EOL) {
      n++;
      if (++ap >= &args[3]) {
	break;
      }
    }
  }
  switch (n) {
    case 3:	/* set HTTP PROTOCOL */
      if (xstrncpy(thd->http.protocol, args[2],
		   sizeof(thd->http.protocol)) == NULL) {
	setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
	return FALSE;
      }
      if (strcasecmp(thd->http.protocol, "HTTP/0.9") == 0) {
	thd->http.version = HTTP_09;
      }
      else if (strcasecmp(thd->http.protocol, "HTTP/1.0") == 0) {
	thd->http.version = HTTP_10;
      }
      else if (strcasecmp(thd->http.protocol, "HTTP/1.1") == 0) {
	thd->http.version = HTTP_11;
#if	BUGGY_KEEP_ALIVE
	if (thd->keepAlive.max > 0) {
	  thd->keepAlive.max--;
	  thd->keepAlive.state = KA_CONTINUE;	/* default in HTTP/1.1 */
#if	DEBUG
	  fprintf(stderr, "[Keep-Alive(%d)]: state = %d\n",
		  thd->client.port, thd->keepAlive.state);
#endif	/* DEBUG */
	}
#endif	/* BUGGY_KEEP_ALIVE */
      }
      else {
	setResponseCode(thd, CODE_HTTP_VERSION_NOT_SUPPORTED);
	return FALSE;
      }
    case 2:	/* set URI */
      if (xstrncpy(thd->requestURI, args[1], sizeof(thd->requestURI)) == NULL) {
	setResponseCode(thd, CODE_REQUEST_URI_TOO_LONG);
	return FALSE;
      }
    case 1:	/* set METHOD */
      if (xstrncpy(thd->method.name, args[0],
		   sizeof(thd->method.name)) == NULL) {
	setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
	return FALSE;
      }
      break;
    default:
      setResponseCode(thd, CODE_BAD_REQUEST);
      return FALSE;
  }
#if	DEBUG
  fprintf(stderr, "[Request-Line(%d)]: \"%s %s %s\"\n",
	  n, thd->method.name, thd->requestURI, thd->http.protocol);
#endif	/* DEBUG */

  /* invalid request-line */
  if (n < 2) {
    setResponseCode(thd, CODE_BAD_REQUEST);
    return FALSE;
  }

  /* check the method */
  if (strcasecmp(thd->method.name, "HEAD") == 0) {
    thd->method.type = M_HEAD;
  }
  else if (strcasecmp(thd->method.name, "GET") == 0) {
    thd->method.type = M_GET;
  }
  else if (strcasecmp(thd->method.name, "POST") == 0) {
    thd->method.type = M_POST;
  }
#if	0	/* Not implemented */
  else if (strcasecmp(thd->method.name, "OPTIONS") == 0) {
    thd->method.type = M_OPTIONS;
  }
  else if (strcasecmp(thd->method.name, "PUT") == 0) {
    thd->method.type = M_PUT;
  }
  else if (strcasecmp(thd->method.name, "DELETE") == 0) {
    thd->method.type = M_DELETE;
  }
  /* local extension method for roaming mode of Netscape Communicator */
  else if (strcasecmp(thd->method.name, "MOVE") == 0) {
    thd->method.type = M_MOVE;
  }
  else if (strcasecmp(thd->method.name, "TRACE") == 0) {
    thd->method.type = M_TRACE;
  }
  else if (strcasecmp(thd->method.name, "CONNECT") == 0) {
    thd->method.type = M_CONNECT;
  }
  /* original extension method for wyvern, but not used... */
  else if (strcasecmp(thd->method.name, "SEARCH") == 0) {
    thd->method.type = M_SEARCH;
  }
#endif
  else {
    return FALSE;	/* Not implemented method */
  }
  return TRUE;
}

/*
 * Function Name:	recvHeader
 * Description:		receive the request header field from a client
 * Arguments:		thread
 * Return Value:	error is FALSE, otherwise is TRUE
 */
static bool_t
recvHeader(thd)
	register Thread *	thd;
{
  int	n;
  char	buf[BUF_SIZE];
  int	if_mod	 = 0;	/* flag for If-Modified-Since/If-Unmodified-Since */
  int	if_match = 0;	/* flag for If-Match/If-Non-Match */

  memset(buf, 0, sizeof(buf));
  while ((n = readline(thd, buf, (size_t)sizeof(buf))) > 0) {
    if (n == -1 || thd->state == IN_ERROR) {	/* an error happened */
      return FALSE;
    }
    else if (strcmp(buf, CRLF) == 0) {	/* the end of request header */
      /*
       * basic error checking:
       * (1) the HTTP/1.1 request need Host field,
       * (2) the conditional request is not overlap,
       * (3) the POST method need Content-Length field.
       */
      if (thd->http.version == HTTP_11 && thd->request.host == NULL) {
	setResponseCode(thd, CODE_BAD_REQUEST);
	return FALSE;
      }
      else if (if_mod > 1 || if_match > 1) {
	setResponseCode(thd, CODE_PRECONDITION_FAILED);
	return FALSE;
      }
      else if (thd->method.type == M_POST && thd->request.content_length == 0) {
	setResponseCode(thd, CODE_LENGTH_REQUIRED);
	return FALSE;
      }
      /*
       * request OK
       */
      else {
	return TRUE;
      }
    }

    /*
     * set request header fields
     */
    chomp(buf);	/* remove CRLF in string */
    if (strncasecmp(buf, "ACCEPT:", 7) == 0) {
      getAcceptField(thd, buf);
    }
    else if (strncasecmp(buf, "ACCEPT-CHARSET:", 15) == 0) {
      thd->request.accept_charset = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "ACCEPT-ENCODING:", 16) == 0) {
      thd->request.accept_encoding = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "ACCEPT-LANGUAGE:", 16) == 0) {
      thd->request.accept_language = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "FROM:", 5) == 0) {
      thd->request.from = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "AUTHORIZATION:", 14) == 0) {
      thd->request.authorization = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "IF-MODIFIED-SINCE:", 18) == 0) {
      thd->request.if_modified_since = getAnyHeaderField(buf);
      if_mod++;
    }
    else if (strncasecmp(buf, "IF-UNMODIFIED-SINCE:", 20) == 0) {
      thd->request.if_unmodified_since = getAnyHeaderField(buf);
      if_mod++;
    }
    else if (strncasecmp(buf, "IF-MATCH:", 9) == 0) {
      thd->request.if_match = getAnyHeaderField(buf);
      if_match++;
    }
    else if (strncasecmp(buf, "IF-NONE-MATCH:", 14) == 0) {
      thd->request.if_none_match = getAnyHeaderField(buf);
      if_match++;
    }
    else if (strncasecmp(buf, "IF-RANGE:", 9) == 0) {
      thd->request.if_range = getAnyHeaderField(buf);
    }
    /* HTTP/1.1 supports persistent connection (default) */
    else if (strncasecmp(buf, "CONNECTION:", 11) == 0) {
      thd->request.connection = getAnyHeaderField(buf);
      if (strncasecmp(thd->request.connection, "keep-alive", 10) == 0) {
#if	BUGGY_KEEP_ALIVE
	if (thd->http.version == HTTP_10 && thd->keepAlive.max > 0) {
	  thd->keepAlive.max--;
	  thd->keepAlive.state = KA_CONTINUE;
	}
#endif	/* BUGGY_KEEP_ALIVE */
      }
      else if (strncasecmp(thd->request.connection, "close", 5) == 0) {
	thd->keepAlive.state = KA_CLOSE;
      }
      else {
	thd->keepAlive.state = KA_DISABLED;
      }
#if	DEBUG
      fprintf(stderr, "[Connection(%d)]: %s[%d] max=%d\n", thd->client.port,
	      thd->request.connection,
	      thd->keepAlive.state, thd->keepAlive.max);
#endif	/* DEBUG */
    }
    else if (strncasecmp(buf, "REFERER:", 8) == 0) {
      thd->request.referer = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "USER-AGENT:", 11) == 0) {
      thd->request.user_agent = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "HOST:", 5) == 0) {
      thd->request.host = getAnyHeaderField(buf);
#if	DEBUG
      if (thd->request.host != NULL) {
	fprintf(stderr, "[Host]: %s\n", thd->request.host);
      }
#endif	/* DEBUG */
    }
    /**/
    else if (strncasecmp(buf, "CONTENT-TYPE:", 13) == 0) {
      thd->request.content_type = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "CONTENT-LENGTH:", 15) == 0) {
      thd->request.content_length = atoi(&buf[15]);
    }
    else if (strncasecmp(buf, "CONTENT-ENCODING:", 17) == 0) {
      thd->request.content_encoding = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "CONTENT-LANGUAGE:", 17) == 0) {
      thd->request.content_language = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "CONTENT-MD5:", 12) == 0) {
      thd->request.content_md5 = getAnyHeaderField(buf);
    }
    /**/
    else if (strncasecmp(buf, "RANGE:", 6) == 0) {
      thd->request.range = getAnyHeaderField(buf);
    }
    /**/
    else if (strncasecmp(buf, "COOKIE:", 7) == 0) {
      thd->request.cookie = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "KEEP-ALIVE:", 11) == 0) {
      ;	/* Gecko sends this request header field ... ingored */
    }
    else if (strncasecmp(buf, "X-FORWARDED-FOR", 15) == 0) {
      thd->request.x_forwarded_for = getAnyHeaderField(buf);
      /*
       * Uncomment the following line if you allow to access
       * from the unknown host which a proxy server forwards.
       */
/*#define	LOGGING_ONLY_X_FORWARDED_FOR_UNKNOWN	1 */
#ifndef	LOGGING_ONLY_X_FORWARDED_FOR_UNKNOWN
     if (strncasecmp(thd->request.x_forwarded_for, "unknown", 7) == 0) {
	setResponseCode(thd, CODE_BAD_REQUEST);
	return FALSE;
     }
#endif	/* LOGGING_ONLY_X_FORWARDED_FOR_UNKNOWN */
    }
#if	SUPPORT_ALL_REQUEST_HEADER
    else if (strncasecmp(buf, "CACHE-CONTROL:", 14) == 0) {
      thd->request.cache_control = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "PRAGMA:", 7) == 0) {
      thd->request.pragma = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "TRAILER:", 8) == 0) {
      thd->request.trailer = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "TRANSFER-ENCODING:", 18) == 0) {
      thd->request.transfer_encoding = getAnyHeaderField(buf);
    }
    /**/
    else if (strncasecmp(buf, "EXPECT:", 7) == 0) {
      thd->request.expect = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "MAX-FORWARDS:", 13) == 0) {
      thd->request.max_forwards = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "PROXY-AUTHORIZATION:", 20) == 0) {
      thd->request.proxy_authorization = getAnyHeaderField(buf);
    }
    else if (strncasecmp(buf, "TE:", 3) == 0) {
      thd->request.te = getAnyHeaderField(buf);
    }
#endif	/* SUPPORT_ALL_REQUEST_HEADER */
    else if (strncasecmp(buf, "VIA:", 4) == 0) {
      thd->request.via = getAnyHeaderField(buf);
    }
    /**/
    else if (strncasecmp(buf, "X-", 2) == 0) {
      /* ignore the other extension header in request-header field */
      wvSyslog(LOG_INFO, "got an extend header field [%s]\n", buf);
#if	DEBUG
      fprintf(stderr, "[header]: get the extend header field [%s]\n", buf);
#endif	/* DEBUG */
    }
    else {
      wvSyslog(LOG_NOTICE, "unexpected header [%s] ... ignored\n", buf);
#if	DEBUG
      fprintf(stderr, "[header]: unexpected header [%s]\n", buf);
#endif	/* DEBUG */
    }
  }

  return TRUE;
}

/*
 * Function Name:	recvEntity
 * Description:		receive the request entity from a client
 * Arguments:		thread
 * Return Value:	none
 */
static void
recvEntity(thd)
	register Thread *	thd;
{
  char	buf[BUF_SIZE];
  int	len;
  int	n = 0;

  /* Wyvern forbids client using POST method except for a CGI access. */
  if (thd->method.type == M_POST && thd->ctype != C_CGI) {
    setResponseCode(thd, CODE_BAD_REQUEST);
    return;
  }

  /* open temporary file */
  thd->cgi.fd = open(thd->cgi.post_tmp, O_RDWR|O_CREAT|O_TRUNC|O_EXCL, 0600);
  if (thd->cgi.fd == -1) {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    return;
  }
  unlink(thd->cgi.post_tmp);	/* the file will be removed when it's closed. */

  /* read data from client */
  for (len = 0; len < thd->request.content_length; len += n) {
    memset(buf, 0, sizeof(buf));
    if ((n = nread(thd, buf, sizeof(buf))) > 0) {
      if (write(thd->cgi.fd, buf, (size_t)n) != n) {
	setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
	return;
      }
    }
  }
#if	DEBUG
  fprintf(stderr, "[entity(%d)]: (%d) %d/%d\n",
	  thd->client.port, n, len, thd->request.content_length);
#endif	/* DEBUG */
  lseek(thd->cgi.fd, (off_t)0, SEEK_SET);	/* rewind file pointer */
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	getAnyHeaderField
 * Description:		get a request/response header filed
 * Arguments:		field
 * Return Value:	pointer
 */
String
getAnyHeaderField(str)
	String	str;
{
  String	value;
  char		c;

  /* move pointer to ':' */
  value = strchr(str, ':');
  if (*++value == EOL) {	/* empty field */
    return NULL;
  }

  while ((c = *value) && (c == SPC || c == TAB)) {
    value++;
  }

  return xstrdup(value);
}


/*
 * Function Name:	getAcceptField
 * Description:		an accept header filed is stored into the buffer
 * Arguments:		thread, accept field
 * Return Value:	pointer
 */
static String
getAcceptField(thd, str)
	register Thread *	thd;
	String			str;
{
  String	value;
  char		c;

  /* move pointer to ':' */
  value = (String)strchr(str, ':');
  if (value++ == NULL) {
    return NULL;
  }
  while ((c = *value) && (c == SPC || c == TAB)) {
    value++;
  }

  if (thd->request.accept == NULL) {
    thd->request.accept = xstrndup(value, strlen(value));
  }
  else {
    String	old = thd->request.accept;
    if (xasprintf(&thd->request.accept, "%s, %s",
		  thd->request.accept, value) == -1) {
      thd->request.accept = old;
    }
    else {
      xfree(old);
    }
  }

  return thd->request.accept;
}

/*
 * Function Name:	checkAcceptField
 * Description:		check accept header filed to response contents
 * Arguments:		thread
 * Return Value:	none
 * Notes:		ignore quality value (ex. q=0.9)
 */
#define	MAX_ACCEPTS	32
static void
checkAcceptField(thd)
	register Thread *	thd;
{
  String	accept;
  size_t	len;
  int		n;
  String	p;
  String *	ap;
  String	args[MAX_ACCEPTS];
  int		i;

  if (thd->request.accept == NULL || thd->mtype == NULL) {
    return;
  }

  accept = xstrdup(thd->request.accept);
  len	 = strcspn(thd->mtype, "/");
  len++;	/* distance to '/' */

  for (n = 0, p = accept, ap = args; (*ap = strsep(&p, ", ;")) != NULL; ) {
    if (**ap != EOL) {
      n++;
      if (++ap >= &args[MAX_ACCEPTS]) {
	break;
      }
    }
  }

  for (i = 0; i < n; i++) {
#if	DEBUG
    fprintf(stderr, "[accept]: %s <= %s\n", thd->mtype, args[i]);
#endif	/* DEBUG */
    if (strcmp(thd->mtype, args[i]) == 0 ||
	(strncmp(thd->mtype, args[i], len) == 0 && args[i][len] == '*') ||
	strcmp(args[i], "*/*") == 0) {
      goto done;	/* OK */
    }
  }

  setResponseCode(thd, CODE_NOT_ACCEPTABLE);
done:
  xfree(accept);
}

