/*****************************************************************************
 *
 * FILE:	ssl.c
 * DESCRIPTION:	secure socket layer module for Wyvern
 * DATE:	Thu, Jan 20 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2000-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ssl.c,v 1.24.2.4 2002/03/07 10:46:18 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <fcntl.h>
#if	HAVE_SYS_FILIO_H
#include <sys/filio.h>
#else
#include <sys/ioctl.h>
#endif	/* HAVE_SYS_FILIO_H */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	HTTPS_SERVICE	"https"
#define	HTTPS_PORT	(443)
#define	SSL_MAX_REQUEST	(5)


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	sslInit(void);
static RSA *	tmpRSACallback(SSL *, int, int);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static SSL_CTX *	ssl_ctx = NULL;
#if	SSLEAY_VERSION_NUMBER >= 0x0922
static unsigned char *	sid_ctx = (unsigned char *)"wyvern SID";
#endif


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	sslInit
 * DESCRIPTION:		initialize SSL
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
sslInit(void)
{
  SSL_METHOD *	mtd = NULL;

#if	!defined(NO_SSL2) && !defined(NO_SSL3)
  mtd = SSLv23_server_method();
#elif	!defined(NO_SSL3)
  mtd = SSLv3_server_method();
#elif	!defined(NO_SSL2)
  mtd = SSLv2_server_method();
#endif

  /* initialize SSL */
  SSL_library_init();		/* same as OpenSSL_add_ssl_algorithms(); */
  SSL_load_error_strings();

  /* create new SSL context */
  ssl_ctx = (SSL_CTX *)SSL_CTX_new(mtd);
  if (ssl_ctx == NULL) {
    die("failed to create a SSL context\n");
  }
#ifndef	NO_RSA
  SSL_CTX_set_tmp_rsa_callback(ssl_ctx, tmpRSACallback);
#endif	/* NO_RSA */

#if	SSLEAY_VERSION_NUMBER >= 0x00906000L
#if	0
  SSL_CTX_set_mode(ssl_ctx, SSL_MODE_AUTO_RETRY);
#endif
  SSL_CTX_set_mode(ssl_ctx, SSL_MODE_ENABLE_PARTIAL_WRITE |
			    SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER |
			    SSL_MODE_AUTO_RETRY);
#endif
  SSL_CTX_set_session_cache_mode(ssl_ctx, SSL_SESS_CACHE_SERVER);
  SSL_CTX_set_timeout(ssl_ctx, 60L);

  /* ceritification */
  if (!SSL_CTX_use_certificate_file(ssl_ctx, WYVERN_PEM, SSL_FILETYPE_PEM)) {
    die("failed to read certificate file: %s\n", WYVERN_PEM);
  }
#if	NO_RSA
  if (!SSL_CTX_use_PrivateKey_file(ssl_ctx, WYVERN_PEM, SSL_FILETYPE_PEM)) {
    die("failed to read PrivateKey file: %s\n", WYVERN_PEM);
  }
#else
  if (!SSL_CTX_use_RSAPrivateKey_file(ssl_ctx, WYVERN_PEM, SSL_FILETYPE_PEM)) {
    die("failed to read RSA-PrivateKey file: %s\n", WYVERN_PEM);
  }
#endif	/* NO_RSA */
  if (!SSL_CTX_check_private_key(ssl_ctx)) {
    die("Private key does not match the certificate\n");
  }

  /* set SSL Context options */
  SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL|SSL_OP_NETSCAPE_CA_DN_BUG);

#if	DEBUG
  fprintf(stderr, "[ssl]: [%s]\n", X509_get_default_cert_dir());
#endif	/* DEBUG */
}


/*
 * FUNCTION NAME:	tmpRSACallback
 * DESCRIPTION:		make a temporary private key
 * ARGUMENTS:		SSL connect, export flag, key length
 * RETURN VALUE:	key 512 / key 1024
 */
static RSA *
tmpRSACallback(ssl_con, export, keylength)
	SSL *	ssl_con;
	int	export;
	int	keylength;
{
  static RSA *	rsa_tmp = NULL;

  if (rsa_tmp == NULL) {
#if	SSLEAY_VERSION_NUMBER >= 0x0900
    rsa_tmp = RSA_generate_key(keylength, RSA_F4, NULL, NULL);
#else
    rsa_tmp = RSA_generate_key(keylength, RSA_F4, NULL);
#endif
  }
  return rsa_tmp;
}

/*
 * FUNCTION NAME:	sslOpenServiceChannel
 * DESCRIPTION:		setup netwok service for SSL
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
sslOpenServiceChannel(void)
{
#if	USE_IPV6
  struct sockaddr_in6	ssl_serv;
#else
  struct sockaddr_in	ssl_serv;
#endif	/* USE_IPV6 */
  int			on;

  sslInit();
#if	USE_IPV6
  httpd->socket.ssl = socket(PF_INET6, SOCK_STREAM, 0);
#else
  httpd->socket.ssl = socket(PF_INET, SOCK_STREAM, 0);
#endif	/* USE_IPV6 */
  if (httpd->socket.ssl == -1) {
    die("failed to create a SSL socket\n");
  }

  /* allow reuse of local addresses */
  on = 1;
  if (setsockopt(httpd->socket.ssl, SOL_SOCKET, SO_REUSEADDR, (void *)&on,
		 sizeof(on)) == -1) {
    die("failed to set a SSL socket option\n");
  }
#if	DEBUG && defined(SO_REUSEPORT)
  if (setsockopt(httpd->socket.ssl, SOL_SOCKET, SO_REUSEPORT, (void *)&on,
		 sizeof(on)) == -1) {
    die("failed to set a SSL socket option\n");
  }
#endif	/* DEBUG && defined(SO_REUSEPORT) */

  /* initialize socket address */
  memset((char *)&ssl_serv, 0, sizeof(ssl_serv));
#if	USE_IPV6
  ssl_serv.sin6_family	= PF_INET6;
  ssl_serv.sin6_port	= htons(httpd->portSSL);
  if (httpd->bindAddress != NULL) {
    if (inet_pton(AF_INET6, httpd->bindAddress, &ssl_serv.sin6_addr) != 1) {
      die("BindAddress[%s] is not valid.\n", httpd->bindAddress);
    }
  }
  else {
    ssl_serv.sin6_addr	= in6addr_any;
  }
#else
  ssl_serv.sin_family	   = PF_INET;
  ssl_serv.sin_port	   = htons(httpd->portSSL);
  if (httpd->bindAddress != NULL) {
    if (inet_pton(AF_INET, httpd->bindAddress, &ssl_serv.sin_addr) != 1) {
      die("BindAddress[%s] is not valid.\n", httpd->bindAddress);
    }
  }
  else {
    ssl_serv.sin_addr.s_addr = htonl(INADDR_ANY);
  }
#endif	/* USE_IPV6 */

  /* bind the socket */
  if (bind(httpd->socket.ssl,
	   (struct sockaddr *)&ssl_serv, sizeof(ssl_serv)) == -1) {
    die("failed to bind a name to a SSL socket\n");
  }

  if (listen(httpd->socket.ssl, SSL_MAX_REQUEST) == -1) {
    die("failed to listen for SSL connection\n");
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	sslCleanup
 * DESCRIPTION:		clean-up SSL
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
sslCleanup(void)
{
  shutdown(httpd->socket.ssl, 2);/* shutdown the full-duplex connection	*/
  xclose(httpd->socket.ssl);	 /* close the SSL socket */
  if (ssl_ctx) {
    SSL_CTX_free(ssl_ctx);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	sslAcceptNewConnection
 * DESCRIPTION:		accept a SSL new connection
 * ARGUMENTS:		thread
 * RETURN VALUE:	Successful: SSL_OK, error: SSL_ERROR, retry: SSL_RETRY
 */
int
sslAcceptNewConnection(thd)
	register Thread *	thd;
{
  thd->SSL.ssl = SSL_new(ssl_ctx);
  if (thd->SSL.ssl == NULL) {
    return SSL_ERROR;
  }
#if	SSLEAY_VERSION_NUMBER >= 0x0922
  SSL_set_session_id_context(thd->SSL.ssl, sid_ctx, strlen(sid_ctx));
#endif
  /* reset SSL object to allow another connection */
  SSL_clear(thd->SSL.ssl);

  /* coonect the SSL object with a socket descriptor */
  SSL_set_fd(thd->SSL.ssl, thd->sd);

  /* initialize the SSL connection to accept state */
  SSL_set_accept_state(thd->SSL.ssl);

  return sslEstablishConnection(thd);
}

/*
 * FUNCTION NAME:	sslEstablishConnection
 * DESCRIPTION:		establish a SSL connection
 * ARGUMENTS:		thread
 * RETURN VALUE:	Successful: SSL_OK, error: SSL_ERROR, retry: SSL_RETRY
 */
ssl_t
sslEstablishConnection(thd)
	register Thread *	thd;
{
  int	n   = 5;	/* counter to retry */
  int	ret = SSL_accept(thd->SSL.ssl);
  switch (SSL_get_error(thd->SSL.ssl, ret)) {
    case SSL_ERROR_NONE:
      thd->state	= RECV_READY;
      thd->SSL.active	= TRUE;
      break;

    case SSL_ERROR_WANT_READ:
      if (n-- > 0) {
	usleep(SSL_USLEEP_TIME);
      }
      thd->state = SSL_READY;
#if	DEBUG
      fprintf(stderr, "[SSL_accept(%d)]: %d/%d/%d more read data = %d\n",
		thd->client.port, SSL_want_read(thd->SSL.ssl),
		BIO_sock_should_retry(ret), BIO_sock_non_fatal_error(errno),
		errno);
#endif	/* DEBUG */
      return SSL_RETRY;

    case SSL_ERROR_ZERO_RETURN:
    case SSL_ERROR_WANT_WRITE:
    case SSL_ERROR_WANT_X509_LOOKUP:
    case SSL_ERROR_SYSCALL:
    case SSL_ERROR_SSL:
    default: 
#if	DEBUG
      fprintf(stderr, "[SSL_accept(%d)]: %d/%d/%d [%d] SSL ERROR\n",
		thd->client.port, SSL_want_read(thd->SSL.ssl),
		BIO_sock_should_retry(ret), BIO_sock_non_fatal_error(errno),
		errno);
#endif	/* DEBUG */
      wvSyslog(LOG_ERR, "failed to accept for SSL connection\n");
      return SSL_ERROR;
  }
  /* improve performance */
  SSL_set_read_ahead(thd->SSL.ssl, 1);

  return SSL_OK;
}

