/*****************************************************************************
 *
 * FILE:	thread.c
 * DESCRIPTION:	manage the thread queues
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: thread.c,v 1.55.2.7 2002/03/07 10:46:18 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static Thread *	freeThread;	/* free thread in thread queue	*/
static Thread *	threadTop;	/* the top of thread queue (const) */
static Thread *	threadQueue;	/* active thread list; top is latest one */
static int	usedThreads;	/* number of used threads */

/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	clearThreadRequestHeaderField(Thread *);
static void	clearThreadResponseHeaderField(Thread *);
static int	gc(void);
static void	initThread(Thread *);
#if	DEBUG
static void	printThread(const String, Thread *);
#endif	/* DEBUG */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	initThreadQueue
 * Description:		create thread queue
 * Arguments:		none
 * Return Value:	none
 */
void
initThreadQueue(void)
{
  register Thread *	thd;
#if	DEBUG
  int			no = 1;
#endif	/* DEBUG */

  /* allocate memory area of threads */
#if	DEBUG_ALLOC
  fprintf(stderr, "[thread]: allocation = %u (%u x %u)\n",
	  (size_t)httpd->maxRequest * sizeof(Thread),
	  (size_t)httpd->maxRequest, sizeof(Thread));
#endif	/* DEBUG_ALLOC */
  freeThread = threadTop =
	(Thread *)xcalloc((size_t)httpd->maxRequest, sizeof(Thread));
  if (freeThread == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }

  /* initialize each thread entity */
  for (thd = threadTop; thd < threadTop + httpd->maxRequest; thd++) {
#if	DEBUG
    thd->no = no++;	/* numbering */
#endif	/* DEBUG */
    initThread(thd);
    cleanupThread(thd);
    if (xasprintf(&thd->cgi.post_tmp, "%s/post%u",
  		  httpd->tmpDir, (unsigned int)thd) == -1) {
      die("POST temporary buffer overflowed in initThread()\n");
    }
    thd->next = thd + 1;
  }
  (--thd)->next = NULL;

  /* initialize thread queue */
  threadQueue = NULL;

  /* initialize number of used threads */
  usedThreads = 0;
}

#if	PARANOIA		/* for free-on-exit */
/*
 * Function Name:	freeThreadQueue
 * Description:		close all connected sockets and detach threads
 * Arguments:		none
 * Return Value:	none
 */
void
freeThreadQueue(void)
{
  XFree(threadTop);
}
#endif	/* PARANOIA */

/*
 * Function Name:	flushThreadQueue
 * Description:		flush all requests in the thread queue
 * Arguments:		none
 * Return Value:	none
 */
void
flushThreadQueue(void)
{
  register Thread *	thd;

  for (thd = threadTop; thd < threadTop + httpd->maxRequest; thd++) {
    if (thd->sd > 0) {
      shutdown(thd->sd, SHUT_RD);	/* receive is disallowed */
      switch (thd->state) {
	case IN_ERROR:
	case SEND_READY:
	  sendResponse(thd);
	case NOW_SENDING:
	  while (thd->state != TERMINATED) {
	    sendingEntity(thd);
	  }
	case TERMINATED:
	  detachThread(thd);
	default:
	  break;
      }
    }
  }
}

/*
 * Function Name:	newThread
 * Description:		get a new thread from thread queue
 * Arguments:		socket descriptor
 * Return Value:	new thread
 */
Thread *
newThread(sd)
	int	sd;
{
  Thread *	new;	/* for new thread */

  if (freeThread == NULL) {
    if (gc() == 0) {	/* garbage collection */
      return NULL;
    }
  }
  new			= freeThread;
  freeThread		= freeThread->next;
  new->sd		= sd;	/* set new socket descriptor */
  new->state		= RECV_READY;
  new->ctype		= C_UNK;
  new->atime		= time(NULL);	/* set last access time */
  new->cgi.envno	= 0;

  /*
   * link a new thread to the thread queue
   */
  if (threadQueue == NULL) {	/* first thread queue */
    new->next  = new->prev	= NULL;
  }
  else {			/* second or later thread queue */
    new->next		= threadQueue;
    new->prev		= threadQueue->prev;
    threadQueue->prev	= new;
  }

  threadQueue = new;
#if	DEBUG
  printThread("new", new);
#endif	/* DEBUG */
  usedThreads++;	/* count up */

  return new;
}

/*
 * Function Name:	detachThread
 * Description:		detach a thread
 * Arguments:		the detached thread
 * Return Value:	none
 */
void
detachThread(thd)
	register Thread *	thd;
{
  thd->state = DETACHED;

  if (thd->prev == NULL && thd->next == NULL) {	/* only one */
    threadQueue = NULL;
#if	DEBUG
    printThread("DETACH(O)", thd);
#endif	/* DEBUG */
  }
  else if (thd->prev == NULL) {	/* first thread */
    thd->next->prev = thd->prev;
    threadQueue	    = thd->next;
#if	DEBUG
    printThread("DETACH(F)", thd);
#endif	/* DEBUG */
  }
  else if (thd->next == NULL) {	/* last thread */
    thd->prev->next = thd->next;
#if	DEBUG
    printThread("DETACH(L)", thd);
#endif	/* DEBUG */
  }
  else {
    thd->prev->next = thd->next;
    thd->next->prev = thd->prev;
#if	DEBUG
    printThread("DETACH(M)", thd);
#endif	/* DEBUG */
  }
  thd->next = thd->prev = NULL;	/* no thread queue */
#if	DEBUG
  fprintf(stderr, "[Thread(%d)]: detached...\n", thd->client.port);
#endif	/* DEBUG */

  /* close connection */
#if	USE_SSL
  if (thd->SSL.ssl != NULL) {
    SSL_shutdown(thd->SSL.ssl);
    SSL_clear(thd->SSL.ssl);
    SSL_free(thd->SSL.ssl);
  }
#endif	/* USE_SSL */
  shutdown(thd->sd, SHUT_RDWR);
  xclose(thd->sd);
  thd->sd = -1;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	getThreadHead
 * Description:		get the head of thread queue
 * Arguments:		none
 * Return Value:	thread header
 */
Thread *
getThreadHead(void)
{
  return threadQueue;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	sweepThreadQueue
 * Description:		only call gc()
 * Arguments:		none
 * Return Value:	none
 */
void
sweepThreadQueue(void)
{
  if (usedThreads > 0) {
    gc();
  }
}

/*
 * Function Name:	gc
 * Description:		collect unused threads
 * Arguments:		none
 * Return Value:	number of collected threads
 */
static int
gc(void)
{
  register Thread *	thd;
  register Thread *	end;
  int			num;

  /* find first free thread */
  for (thd = threadTop;
       thd < threadTop + httpd->maxRequest &&
       thd->state != UNUSED && thd->state != DETACHED; thd++) {
    ;	/* nothing to do */
  }
  if (thd >= threadTop + httpd->maxRequest) {
    return 0;	/* no more free threads */
  }
  if (thd->state == DETACHED) {
    initThread(thd);
    num++;
  }
  freeThread	= end = thd;
  end->prev	= NULL;

  /* find free thread sequentially */
  for (num = 1, ++thd; thd < threadTop + httpd->maxRequest; thd++) {
    if (thd->state == UNUSED || thd->state == DETACHED) {
      if (thd->state == DETACHED) {
	initThread(thd);
	num++;
      }
      end->next	= thd;
      end	= thd;
    }
  }
  end->next = NULL;

#if	DEBUG
  for (thd = freeThread; thd != NULL; thd = thd->next) {
    printThread("garbaged", thd);
  }
  fprintf(stderr, "[gc]: garbaged=%d\n", num);
#endif	/* DEBUG */

  /* initialize number of used threads */
  usedThreads = 0;

  return num;
}

/*
 * Function Name:	initThread
 * Description:		initialize a thread's members
 * Arguments:		thread
 * Return Value:	none
 */
static void
initThread(thd)
	register Thread *	thd;
{
  thd->sd		= -1;

  thd->ios.update	= FALSE;
  thd->ios.total	= 0;
  thd->ios.next		= 0;
  memset(thd->ios.buf, 0, sizeof(thd->ios.buf));

  thd->client.port	= 0;
  XFree(thd->client.name);
  XFree(thd->client.ip);

  thd->keepAlive.state	= KA_DISABLED;
  thd->keepAlive.max	= httpd->ka_max;
  thd->keepAlive.timeout= httpd->ka_timeout;
  thd->atime		= 0;

  XFree(thd->userDir);
  thd->userHome = NULL;	/* this pointer indicates the value of getHomeDir() */
  XFree(thd->userName);
  XFree(thd->identUser);
  XFree(thd->authUser);
  memset((char *)&thd->file, 0, sizeof(thd->file));

  thd->http.version	= HTTP_10;
  memset(thd->http.protocol, 0, sizeof(thd->http.protocol));
  strncpy(thd->http.protocol, HTTP_VERSION, sizeof(thd->http.protocol) - 1);
#if	USE_SSL
  thd->SSL.active	= FALSE;
  thd->SSL.ssl		= NULL;
#endif	/* USE_SSL */
  thd->next =thd->prev = NULL;
#if	DEBUG
  printThread("initThread", thd);
#endif	/* DEBUG */
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	cleanupThread
 * Description:		cleanup partial varibales in a thread structure,
 *			and close file descriptors;
 *			not cleanup variables for network
 * Arguments:		thread
 * Return Value:	none
 */
void
cleanupThread(thd)
	register Thread *	thd;
{
  thd->state		= UNUSED;
  thd->rsCode		= CODE_REQUEST_OK;
  thd->mtype		= NULL;
  thd->ctype		= C_UNK;

  thd->range.first	= 0;
  thd->range.last	= 0;
  thd->parts.fsize	= 0;
  thd->parts.bsize	= 0;
  thd->parts.sum	= 0;

  /* clear request */
  thd->method.type	= M_UNKNOWN;
  memset(thd->method.name, 0, sizeof(thd->method.name));
  memset(thd->requestURI,  0, sizeof(thd->requestURI));
  XFree(thd->path);

  /* free request header fields */
  clearThreadRequestHeaderField(thd);

  /* free response header fields */
  clearThreadResponseHeaderField(thd);

  /* close file pointer */
  if (thd->fp != NULL) {
    XFClose(thd->fp);
  }

  /* close temporary file descriptor */
  if (thd->cgi.fd > 0) {
    xclose(thd->cgi.fd);
    thd->cgi.fd = -1;
  }
  /* kill child process (by time over) */
  if (thd->cgi.pid > 0) {
    kill(thd->cgi.pid, SIGTERM);
    wvSyslog(LOG_ERR, "killed CGI process %d", thd->cgi.pid);
    thd->cgi.pid = -1;
  }
  XFree(thd->cgi.path_info);
  XFree(thd->cgi.query_string);
  XFree(thd->cgi.status);
  /* free header list */
  if (thd->cgi.headers != NULL) {
    register StrList *	slp;
    register StrList *	tmp;
    for (slp = thd->cgi.headers; slp != NULL; slp = tmp) {
      tmp = slp->next;
      XFree(slp->text);
      XFree(slp);
    }
  }
  thd->cgi.headers = NULL;
  /* free environment */
  if (thd->cgi.env != NULL) {
    register int	i;

    for (i = httpd->envno; i < thd->cgi.envno && thd->cgi.env[i] != NULL; i++) {
      XFree(thd->cgi.env[i]);
    }
    XFree(thd->cgi.env);
  }
  thd->cgi.envno = 0;
  thd->cgi.env	 = NULL;
#if	HAVE_ZLIB_H
  thd->gzippct	 = 0.0;
#endif	/* HAVE_ZLIB_H */
}

/*
 * Function Name:	clearThreadRequestHeaderField
 * Description:		release allocated memory for request header fields
 * Arguments:		thread
 * Return Value:	none
 */
static void
clearThreadRequestHeaderField(thd)
	register Thread *	thd;
{
  XFree(thd->request.connection);
  XFree(thd->request.via);
#if	SUPPORT_ALL_REQUEST_HEADER
  XFree(thd->request.cache_control);
  XFree(thd->request.pragma);
  XFree(thd->request.trailer);
  XFree(thd->request.transfer_encoding);
#endif	/* SUPPORT_ALL_REQUEST_HEADER */

  XFree(thd->request.accept);
  XFree(thd->request.accept_charset);
  XFree(thd->request.accept_encoding);
  XFree(thd->request.accept_language);
  XFree(thd->request.authorization);
  XFree(thd->request.from);
  XFree(thd->request.host);
  XFree(thd->request.if_match);
  XFree(thd->request.if_modified_since);
  XFree(thd->request.if_none_match);
  XFree(thd->request.if_range);
  XFree(thd->request.if_unmodified_since);
  XFree(thd->request.range);
  XFree(thd->request.referer);
  XFree(thd->request.user_agent);
#if	SUPPORT_ALL_REQUEST_HEADER
  XFree(thd->request.expect);
  XFree(thd->request.max_forwards);
  XFree(thd->request.proxy_authorization);
  XFree(thd->request.te);
#endif	/* SUPPORT_ALL_REQUEST_HEADER */

  XFree(thd->request.content_encoding);
  XFree(thd->request.content_language);
  thd->request.content_length = 0;
  XFree(thd->request.content_md5);
  XFree(thd->request.content_type);
  XFree(thd->request.cookie);
  XFree(thd->request.x_forwarded_for);
}

/*
 * Function Name:	clearThreadResponseHeaderField
 * Description:		release allocated memory for response header fields
 * Arguments:		thread
 * Return Value:	none
 */
static void
clearThreadResponseHeaderField(thd)
	register Thread *	thd;
{
  XFree(thd->response.connection);
  XFree(thd->response.date);
#if	SUPPORT_ALL_RESPONSE_HEADER
  XFree(thd->response.trailer);
  XFree(thd->response.transfer_encoding);
  XFree(thd->response.upgrade);
  XFree(thd->response.via);
  XFree(thd->response.warning);
#endif	/* SUPPORT_ALL_RESPONSE_HEADER */

  XFree(thd->response.accept_ranges);
  XFree(thd->response.etag);
  XFree(thd->response.location);
  XFree(thd->response.server);
  XFree(thd->response.www_authenticate);
#if	SUPPORT_ALL_RESPONSE_HEADER
  XFree(thd->response.age);
  XFree(thd->response.proxy_authenticate);
  XFree(thd->response.retry_after);
  XFree(thd->response.vary);
#endif	/* SUPPORT_ALL_RESPONSE_HEADER */

  XFree(thd->response.allow);
  XFree(thd->response.last_modified);
  XFree(thd->response.content_encoding);
  thd->response.content_length = 0;
  XFree(thd->response.content_location);
  XFree(thd->response.content_md5);
  XFree(thd->response.content_range);
  XFree(thd->response.content_type);
#if	SUPPORT_ALL_RESPONSE_HEADER
  XFree(thd->response.content_language);
  XFree(thd->response.expires);
#endif	/* SUPPORT_ALL_RESPONSE_HEADER */
  XFree(thd->response.set_cookie);
}


#if	DEBUG
static void
printThread(label, thd)
	const String		label;
	register Thread *	thd;
{
  fprintf(stderr, "[%s]: Thread[%d]", label, thd->no);
  if (thd->next != NULL) {
    fprintf(stderr, " Next[%d]", thd->next->no);
  }
  if (thd->prev != NULL) {
    fprintf(stderr, " Prev[%d]", thd->prev->no);
  }
  fputs("\n", stderr);
}
#endif	/* DEBUG */

